/*******************************************************************
cop
---
written by Stefan Beck between 09/92 and 06/94 (puuhh)

changed by Gabriel De Dominicis, starting 09/95

*******************************************************************/


#include <Rational.h>


#ifndef Fp_cc
  #include "Fp.h"
  typedef Fp K_Fp;
#endif

#ifndef Fp8_cc
  #include "Fp8.h"
  #ifndef Fp_cc 
    #include "Fp.h"
  #endif
  typedef Fp8 K_FP8;
#endif


#ifndef QN_cc
  #include "QN.h"
  typedef QN K_QN;
  extern QN *QN_phi;
#endif


#ifdef field_Fp8
typedef Fp8 K;
#endif
#ifdef field_Fp
typedef Fp K;
#endif
#ifdef field_QN
typedef QN K;
#endif
#ifdef field_Q
typedef Rational K;
#endif



#include "monom.h"
#include <procbuf.h>
#include <new.h>
#include <iostream.h>
#include <fstream.h>
#include <ctype.h>
#include <string.h>
#include <minmax.h>
#include <builtin.h>
#include <ACG.h>
#include <RndInt.h>

int counter;
int LL;
int range1, range2;
int monom_ordering;
int dim;
int proj;
int verbose;
char varstring[30] = "XYZTabcdefghijklmnopqrstuvw";
char Line[255] =
"-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n";
int (*monom_ordering_lt)(monom &x, monom &y);
int (*monom_ordering_le)(monom &x, monom &y);
int (*proj_monom_ordering_lt)(monom &x, monom &y);
int (*proj_monom_ordering_le)(monom &x, monom &y);

ACG gen(1234);


void mem_warn() {
	cerr << "\nNew: can't allocate enough memory\n";
	exit(10);
};

/*******************************************************************
error
*******************************************************************/

void error(int err) {
	switch (err) {
		case 30 :
			cout << "can't open file\n";

		case 31 :
			cout << "can't read scheme because of different dimensions\n";

		case 102 :
			cout << "command or indentifier unknown\n";
			break;

		case 104 :
			cout << "command needs indentifier\n";
			break;

		case 105 :
			cout << "indentifiers have to consist only of alnums\n";
			break;

		case 107 :
			cout << "indentifier unknown\n";
			break;

		case 108 :
			cout << "indentifier of wrong typ\n";
			break;

		case 110 :
			cout << "identifier memory exceeded\n";
			break;

		case 111 :
			cout << "indentifier already exists\n";
			break;
	};
	return;
};


/*******************************************************************
my_func
*******************************************************************/

template <class K>
void swap(K &a, K &b) {
	K tmp = b;
	b = a;
	a = tmp;
	return;
};

int len(int x) {
	int s = 0;
	if(x < 0) {
		s = 1;
		x = -x;
	};
	if(x >= 10000)
		return 5 + s;
	if(x >= 1000)
		return 4 + s;
	if(x >= 100)
		return 3 + s;
	if(x >= 10)
		return 2 + s;
	return 1 + s;
};

int len(Rational x) {
	if (sign(x) == 0)
		return 1;
	int res = 0;
	if (sign(x) < 0) {
		res = 1;
		x.negate();
	};
	Integer n = x.numerator();
	char *a = Itoa(n);
	res += strlen(a);
	delete[] a;
	n = x.denominator();
	if (n == 1)
		return res;
	res++; //  x/y
	a = Itoa(n);
	res += strlen(a);
	delete[] a;
	return res;
};

/*******************************************************************
arithmetic for a polynomialring over a ring or field : K[X,Y,..]
*******************************************************************/

struct polynomial {
	// data
	int   no_of_monoms;     // number of monoms
	monom *term;            // monoms
	K     *coef;            // coefs
	// constructors
	polynomial (polynomial &f);
	polynomial (monom &m);
	polynomial (int x = 0);
	// destructor
	~polynomial ();
	// elements
	polynomial& operator = (polynomial &x);
	polynomial& operator = (monom &x);
	polynomial& operator = (int x);
	polynomial& operator += (polynomial &x);
	polynomial& operator -= (polynomial &x);
	polynomial& operator *= (K &x);
	polynomial& operator *= (monom &x);
	// friends
	friend polynomial operator + (polynomial &f, polynomial &g);
	friend polynomial operator - (polynomial &f, polynomial &g);
	friend polynomial operator * (polynomial &f, polynomial &g);
	friend polynomial operator * (polynomial &f, K &x);
	friend polynomial operator / (polynomial f, polynomial &g);
	friend polynomial operator + (polynomial &f);
	friend polynomial operator - (polynomial f);
	friend polynomial operator / (polynomial &f, monom &m);
// friend int operator < (polynomial &f, polynomial &g);
// friend int operator > (polynomial &f, polynomial &g);
	friend int operator == (polynomial &f, polynomial &g);
	friend ostream& operator << (ostream &s, polynomial &f);
	friend polynomial Lf(polynomial &);
	void put(ostream &s);
};

/*******************************************************************
constructors
*******************************************************************/

polynomial::polynomial(polynomial &f) {
	term = new monom[no_of_monoms = f.no_of_monoms];
	coef = new K[no_of_monoms];
	for(int i = 0; i < no_of_monoms; i++) {
		term[i] = f.term[i];
		coef[i] = f.coef[i];
	};
};

polynomial::polynomial(monom &m) {
	term = new monom[no_of_monoms = 1];
	coef = new K[1];
	term[0] = m;
	coef[0] = 1;
};

polynomial::polynomial(int x) {
	no_of_monoms = 1;
	term = new monom[1];
	coef = new K[1];
	term[0] = 0;
	coef[0] = x;
};

void polynomial::put(ostream &s) {
	for(int i = 0; i < no_of_monoms; i++) {
		if (sign(coef[0]) < 0 && i == 0)
			s << "-";
		if (abs(coef[i]) != 1 || sign(term[i]) == 0)
			s << abs(coef[i]);
		if (sign(term[i]) != 0)
			term[i].put(s);
		if (i != no_of_monoms - 1) {
			if (sign(coef[i + 1]) > 0)
				s << "+";
			else
				s << "-";
		};
	};
};

/*******************************************************************
destructor
*******************************************************************/

polynomial::~polynomial() {
	delete[] term;
	delete[] coef;
};

/*******************************************************************
assign ops
*******************************************************************/

polynomial& polynomial::operator = (polynomial &f) {
	if (this != &f) { // f = f ?
		if(no_of_monoms < f.no_of_monoms) {
			delete[] term;
			delete[] coef;
			term = new monom[f.no_of_monoms];
			coef = new K[f.no_of_monoms];
		};
		no_of_monoms = f.no_of_monoms;
		for(int i = 0; i < no_of_monoms; i++) {
			term[i] = f.term[i];
			coef[i] = f.coef[i];
		};
	};
	return *this;
};

polynomial& polynomial::operator = (monom &f) {
	term[0] = f;
	coef[0] = 1;
	no_of_monoms = 1;
	return *this;
};

polynomial& polynomial::operator = (int x) {
	term[0] = 0;
	coef[0] = x;
	no_of_monoms = 1;
	return *this;
};

/*******************************************************************
relops
*******************************************************************/

int operator < (polynomial &f, polynomial &g) {
	return (f.term[0] < g.term[0]);
};

int operator > (polynomial &f, polynomial &g) {
	return (f.term[0] > g.term[0]);
};

int operator == (polynomial &f, polynomial &g) {
	if (f.no_of_monoms != g.no_of_monoms)
		return 0;
	for(int i = 0; i < f.no_of_monoms; i++)
		if (f.term[i] != g.term[i] || f.coef[i] != g.coef[i])
			return 0;
	return 1;
};

/*******************************************************************
unary ops
*******************************************************************/

polynomial operator + (polynomial &f) {
	return f;
};

polynomial operator - (polynomial f) {
	for(int i = 0; i < f.no_of_monoms; i++)
		f.coef[i] = - f.coef[i];
	return f;
};

/*******************************************************************
binary ops
*******************************************************************/

polynomial operator + (polynomial &f, polynomial &g) {
	polynomial tmp = f;
	return tmp += g;
};

polynomial operator - (polynomial &f, polynomial &g) {
	polynomial tmp = f;
	return tmp -= g;
};

polynomial operator * (polynomial &f, K &x) {
	if(sign(x) == 0)
		return polynomial(0);
	polynomial tmp = f;
	for (int i = 0; i < f.no_of_monoms; i++)
		tmp.coef[i] = f.coef[i] * x;
	return tmp;
};

polynomial& polynomial::operator += (polynomial &g) {
	if (sign(g.coef[0]) == 0)
		return *this;
	if (sign(coef[0]) == 0) {
		if (g.no_of_monoms > 1) {
			delete[] term;
			delete[] coef;
			term = new monom[g.no_of_monoms];
			coef = new K[g.no_of_monoms];
		};
		no_of_monoms = g.no_of_monoms;
		for(int i = 0; i < no_of_monoms; i++) {
			term[i] = g.term[i];
			coef[i] = g.coef[i];
		};
		return *this;
	};
	int l, n;
	l = no_of_monoms;
	n = g.no_of_monoms;
	monom *oldterm;
	K     *oldcoef;
	oldterm = term;
	oldcoef = coef;
	coef = new K[l + n];
	term = new monom[l + n];
	no_of_monoms = 0;
	int i = 0, k = 0;
	while (i < l && k < n) {
		if (monom_ordering_lt(oldterm[i], g.term[k])) {
			term[no_of_monoms] = g.term[k];
			coef[no_of_monoms++] = g.coef[k++];
		}
		else if (monom_ordering_lt(g.term[k], oldterm[i])) {
			term[no_of_monoms] = oldterm[i];
			coef[no_of_monoms++] = oldcoef[i++];
		}
		else {
			coef[no_of_monoms] = oldcoef[i++] + g.coef[k++];
			if (sign(coef[no_of_monoms]) != 0)
				term[no_of_monoms++] = oldterm[i - 1];
		};
	};
	for (; i < l; i++) {
		term[no_of_monoms] = oldterm[i];
		coef[no_of_monoms++] = oldcoef[i];
	};
	for (; k < n; k++) {
		term[no_of_monoms] = g.term[k];
		coef[no_of_monoms++] = g.coef[k];
	};
	delete[] oldcoef;
	delete[] oldterm;
	if (sign(coef[0]) == 0) {
		term[0] = 0;
		no_of_monoms = 1;
	};
// cerr << " " << no_of_monoms << " ";
	return *this;
};

polynomial& polynomial::operator -= (polynomial &g) {
	if (sign(g.coef[0]) == 0)
		return *this;
	if (sign(coef[0]) == 0) {
		if (g.no_of_monoms > 1) {
			delete[] term;
			delete[] coef;
			term = new monom[g.no_of_monoms];
			coef = new K[g.no_of_monoms];
		};
		no_of_monoms = g.no_of_monoms;
		for(int i = 0; i < no_of_monoms; i++) {
			term[i] = g.term[i];
			coef[i] = - g.coef[i];
		};
		return *this;
	};
	int l, n;
	l = no_of_monoms;
	n = g.no_of_monoms;
	monom *oldterm;
	K     *oldcoef;
	oldterm = term;
	oldcoef = coef;
	coef = new K[l + n];
	term = new monom[l + n];
	no_of_monoms = 0;
	int i = 0, k = 0;
	while (i < l && k < n) {
		if (monom_ordering_lt(oldterm[i], g.term[k])) {
			term[no_of_monoms] = g.term[k];
			coef[no_of_monoms++] = -g.coef[k++];
		}
		else if (monom_ordering_lt(g.term[k], oldterm[i])) {
			term[no_of_monoms] = oldterm[i];
			coef[no_of_monoms++] = oldcoef[i++];
		}
		else {
			coef[no_of_monoms] = oldcoef[i++] - g.coef[k++];
			if (sign(coef[no_of_monoms]) != 0)
				term[no_of_monoms++] = oldterm[i - 1];
		};
	};
	for (; i < l; i++) {
		term[no_of_monoms] = oldterm[i];
		coef[no_of_monoms++] = oldcoef[i];
	};
	for (; k < n; k++) {
		term[no_of_monoms] = g.term[k];
		coef[no_of_monoms++] = - g.coef[k];
	};
	delete[] oldcoef;
	delete[] oldterm;
	if (sign(coef[0]) == 0) {
		term[0] = 0;
		no_of_monoms = 1;
	};
// cerr << " " << no_of_monoms << " ";
	return *this;
};

polynomial& polynomial::operator *= (K &x) {
	if(sign(x) == 0) {
		*this = 0;
		return *this;
	};
	for (int i = 0; i < no_of_monoms; i++)
		coef[i] *= x;
	return *this;
};

int sign(polynomial &f);

polynomial& polynomial::operator *= (monom &x) {
	if(sign(*this) == 0)
		return *this;
	for (int i = 0; i < no_of_monoms; i++)
		term[i] *= x;
	return *this;
};


polynomial operator * (polynomial &f, polynomial &g) {
	if (sign(f) == 0)
		return polynomial(0);
	if (sign(g) == 0)
		return polynomial(0);
	polynomial res = 0, tmp;
	for(int i = 0; i < g.no_of_monoms; i++) {
		tmp = f;
		tmp *= g.term[i];
		tmp *= g.coef[i];
		res += tmp;
	};
	return res;
};

/*******************************************************************
I/O
*******************************************************************/

ostream& operator << (ostream &s, polynomial &f) {
	int d;
	if(sign(f) == 0) {
		s << " 0";
		counter += 2;
		return s;
	};
	for(int i = 0; i < f.no_of_monoms; i++) {
		d = 0;
		if (sign(f.coef[i]) > 0) {
			if(i != 0) d += 3;
		}
		else {
			if(i == 0) d += 2;
			else       d += 3;
		};
		if (abs(f.coef[i]) != 1 || sign(f.term[i]) == 0)
			d += len(abs(f.coef[i]));
		if (sign(f.term[i]) != 0)
			d += len(f.term[i]);

		if (counter + d >= LL - 1) {
			if(i == 0)
				counter += d;
			else {
				s << "\n";
				counter = d;
			};
		}
		else counter += d;

		if (sign(f.coef[i]) > 0) {
			if(i != 0) s << " + ";
		}
		else {
			if(i == 0) s << "- ";
			else       s << " - ";
		};

		if (abs(f.coef[i]) != 1 || sign(f.term[i]) == 0)
			s << abs(f.coef[i]);
		if (sign(f.term[i]) != 0)
			s << f.term[i];
	};
	return s;
};

/*******************************************************************
common
*******************************************************************/

int sign(polynomial &f) {
	return abs(sign(f.coef[0]));
};

int is_prj(polynomial &f) {
	return f.term[0].exp[proj - 1];
};

polynomial Lf(polynomial &f) {
	polynomial tmp, tmp1;
	if(sign(f) == 0)
		return 0;
	int deg = f.term[0].tot_deg();
	for (int i = 0; i < f.no_of_monoms; i++) {
		if (f.term[i].tot_deg() != deg)
			break;
		else {
			tmp1 = f.term[i];
			tmp1 *= f.coef[i];
			tmp += tmp1;
		};
	};
	return tmp;
};

void hom(polynomial &f, polynomial &g) {
	int d = dim;
	dim = proj;
	int i, j;
	f.no_of_monoms = g.no_of_monoms;
	delete[] f.term;
	delete[] f.coef;
	f.term = new monom[f.no_of_monoms];
	f.coef = new K[f.no_of_monoms];
	for(i = 0; i < f.no_of_monoms; i++) {
		f.coef[i] = g.coef[i];
		for(j = 0; j < dim - 1; j++)
			f.term[i].exp[j] = g.term[i].exp[j];
		f.term[i].exp[dim - 1] = (char)(g.term[0].tot_deg() - g.term[i].tot_deg());
	};
	dim = d;
};

/*******************************************************************
scheme
*******************************************************************/

template <class R>
struct part_of_list {
	part_of_list *next;
	R item;
};

template <class R>
struct list {
	// data
	int no_of_items;
	part_of_list<R> *first;
	part_of_list<R> *last;
	// constuctors
	list ();
	// destructor
	~list ();
	// elements
	void add(R &x);
	void add_set(R &x);
	void min_del(R &x);
	int is_element(R &x);
};

/*******************************************************************
constructors
*******************************************************************/

template <class R>
list<R>::list() { // creates empty set
	no_of_items = 0;
	first = last = 0;
};

/*******************************************************************
destructor
*******************************************************************/

template <class R>
list<R>::~list () {
	part_of_list<R> *ptr;
	if (no_of_items == 0)
		return;
	ptr = first->next;
	while (first != 0) {
		delete first;
		first = ptr;
		ptr = ptr->next;
	};
};

/*******************************************************************
element functions
*******************************************************************/

template <class R>
int list<R>::is_element (R &x) {
	part_of_list<R> *ptr;
	if (no_of_items == 0)
		return 0;
	ptr = first;
	while(ptr != last) {
		if (ptr->item == x)
			return 1;
		ptr = ptr->next;
	};
	return (ptr->item == x);
};

template <class R>
void list<R>::add (R &x) {
	if (no_of_items == 0) {
		no_of_items = 1;
		first = new part_of_list<R>;
		last = first;
		first->item = x;
		first->next = 0;
	}
	else {
		no_of_items++;
		last->next = new part_of_list<R>;
		last = last->next;
		last->next = 0;
		last->item = x;
	};
};

template <class R>
void list<R>::add_set (R &x) {
	if(is_element(x))
		return;
	add(x);
};

template <class R>    // no check for list is empty
void list<R>::min_del (R &x) {
	part_of_list<R> *ptr, *ptr2;
// just one item
	if (no_of_items == 1) {
		x = first->item;
		delete first;
		no_of_items = 0;
		first = last = 0;
		return;
	};
	no_of_items--;
	ptr  = first;
	ptr2 = first;
	while(ptr->next != 0) {
		if (ptr->next->item < ptr2->next->item)
			ptr2 = ptr;
		ptr = ptr->next;
	};
	ptr = ptr2->next;
	if (ptr->item > first->item) {
		x = first->item;
		ptr = first->next;
		delete first;
		first = ptr;
		return;
	};
	x = (ptr2->next)->item;
	ptr2->next = ptr->next;
	delete ptr;
	if (ptr2->next == 0)
		last = ptr2;
};

/*******************************************************************
I/O
*******************************************************************/

template <class R>
ostream& operator << (ostream &s, list<R> &x) {
	part_of_list<R> *ptr;
	ptr = x.first;
// s << "number of items: " << x.no_of_items << "\n";
	while (ptr != 0) {
		s << ptr->item;
		ptr = ptr->next;
	};
	return s;
};


struct mmm {
	monom t;
	int   n;
	int   k;
	int ins;
	mmm() : t() {};
	~mmm(){};
	mmm& operator = (mmm &m);
	friend int operator < (mmm &m1, mmm &m2);
	friend int operator > (mmm &m1, mmm &m2);
	friend int operator == (mmm &m1, mmm &m2);
};

mmm& mmm::operator = (mmm &m) {
	t = m.t;
	n = m.n;
	k = m.k; // t = O[n] * X_k;
	ins = m.ins;
	return *this;
};

int operator < (mmm &m1, mmm &m2) {
		return (m1.t < m2.t);
};

int operator > (mmm &m1, mmm &m2) {
		return (m1.t > m2.t);
};

int operator == (mmm &m1, mmm &m2) {
		return (m1.t == m2.t);
};


#define no_of_vars 50

struct var_struct {
	char name[13];       // 12 alnum's & \0
	int  typ;            // type of first parameter
	void *var_ptr;       // ptr to value
};

var_struct var[no_of_vars];

#define no_of_cmds 67

struct command_struct {
	char name[13];       // 12 alnum's & \0
	int  par1_typ;       // type of first parameter
};

command_struct command[no_of_cmds] = {
{  "alpha",          3  },{   "betamat",        3  },
{  "bumo",           3  },{   "char",           -1 },
{  "cyclotom",       -1 },{   "cyclotom2",      -1 },
{  "cyclpol",        -1 },{   "deglex",         -1 },
{  "degrevlex",      -1 },{   "del",            0  },
{  "del_all",        -1 },{   "delta",          3  },
{  "deltafn",        3  },{   "dim",            -1 },
{  "display",        0  },{   "display_all",    -1 },
{  "exit",           -1 },{   "gbasis",         3  },
{  "gencycl",        -1 },{   "get",            -1 },
{  "help",           -1 },{   "hilbertfn",      3  },
{  "identifiers",    -1 },{   "is_cb",          3  },
{  "is_duf",         3  },{   "is_gor",         3  },
{  "is_uf",          3  },{   "lingenpos",      3  },
{  "linid",          3  },{   "linom",          3  },
{  "linomsyz",       3  },{   "linres",         3  },
{  "linsyz",         3  },{   "list",           -1 },
{  "min",            3  },{   "mindegs",        3  },
{  "omega",          3  },{   "omegamat",       3  },
{  "orderideal",     3  },{   "putbeta",        3  },
{  "putgbasis",      3  },{   "putlinid",       3  },
{  "putlinom",       3  },{   "putlinomsyz",    3  },
{  "putlinres",      3  },{   "putlinsyz",      3  },
{  "putmin",         3  },{   "putomega",       3  },
{  "putscheme",      3  },{   "putsep",         3  },
{  "quiet",          -1 },{   "quit",           -1 },
{  "random",         -1 },{   "random2",        -1 },
{  "range",          -1 },{   "ring",           -1 },
{  "root",           -1 },{   "scheme",         -1 },
{  "sep",            3  },{   "sigma",          3  },
{  "socle",          3  },{   "type",           3  },
{  "verbose",        -1 },{   "write",          3  },
{  "writelinid",     3  },{   "writelinom",     3  },
{  "writelinres",    3  } };

/*******************************************************************
matrix
*******************************************************************/

struct matrix {
	//data
	int cols;               // number of colomns
	int rows;               // number of rows
	K *(*a);
	// constructors
	matrix (matrix &m);
	matrix (int c = 1, int r = 1);
	// destructor
	~matrix ();
	// elements
	void init();
	K minor(int n, int *col, int *row);
	int solve(int *col, int *row);   // returns dim of solution
									// c contains the col_permutation
									// r contains the row_permutation
	// friends
	friend ostream& operator << (ostream &s, matrix &m);
	friend istream& operator >> (istream &s, matrix &m);

};

/*******************************************************************
constructors
*******************************************************************/

matrix::matrix (matrix &m) {
	cols = m.cols;
	rows = m.rows;
	int i, j;
	a = new K*[rows];
	for(i = 0; i < rows; i++) {
		a[i] = new K[cols];
		for(j = 0; j < cols; j++)
			a[i][j] = m.a[i][j];
	};
};

matrix::matrix (int c, int r) {
	cols = c;
	rows = r;
	a = new K*[rows];
	for(int i = 0; i < rows; i++)
		a[i] = new K[cols];
};

/*******************************************************************
destructors
*******************************************************************/

matrix::~matrix () {
	for(int i = 0; i < rows - 1; i++)
		delete[] a[i];
	delete[] a;
};

/*******************************************************************
elements
*******************************************************************/

void matrix::init() {
	int i, j;
	for(i = 0; i < rows; i++)
		for(j = 0; j < cols; j++)
			a[i][j] = 0;
};

K matrix::minor(int n, int *col, int *rr) {
	// cholesky method
	int i, j, k, m, ii;
	K tmp2;
	int tmp1 = 1;

	// rescue rr
	int *row = new int[n];
	for (i = 0; i < n; i++) {
		row[i] = rr[i];
	};

	if(sign(a[row[0]][col[0]]) == 0) { // pivot-choice for first col
		for(i = 1; i < n; i++)
			if(sign(a[row[i]][col[0]]) != 0) {
				swap(row[i], row[0]);
				tmp1 = -tmp1;
				break;
			};
		if(i == n)
			return K(0);
	};

	matrix mm(n, n);


	for(i = 0; i < n; i++)
		mm.a[i][0] = a[row[i]][col[0]];

	for(j = 1; j < n; j++)
		mm.a[0][j] = a[row[0]][col[j]] / mm.a[0][0]; //step 2
	for(m = 1; m < n; m++) {
		// treat case m,m here separately
		for(ii = m; ii < n; ii++) {   // col-pivot-choice
			tmp2 = 0;
			for(k = 0; k < m; k++)
				tmp2 += mm.a[ii][k] * mm.a[k][m];
			if(tmp2 != a[row[ii]][col[m]]) {
				if(ii != m) {
					swap(row[ii], row[m]);
					tmp1    = -tmp1;
					swap(mm.a[ii], mm.a[m]);
				};
				mm.a[m][m] = a[row[m]][col[m]] - tmp2;
				break;
			};
		}; // end ii

		if (ii == n)
			return K(0);

		for(i = m + 1; i < n; i++) {
			mm.a[i][m] = a[row[i]][col[m]];
			for(k = 0; k < m; k++)
				mm.a[i][m] -= mm.a[i][k] * mm.a[k][m]; // step 3
		};

		for(j = m + 1; j < n; j++) {
			mm.a[m][j] = a[row[m]][col[j]];
			for(k = 0; k < m; k++)
				mm.a[m][j] -= mm.a[m][k] * mm.a[k][j]; // step 4
		mm.a[m][j] /= mm.a[m][m];
		};
	};

	tmp2 = mm.a[0][0];
	for(i = 1; i < n; i++)
		tmp2 *= mm.a[i][i];

	delete[] row;

	if (tmp1 < 0)
		return -tmp2;
	return tmp2;
};

int matrix::solve(int *c, int *r) {
	int i, j, k;
	int col_counter = 0;
	int n = min(cols, rows);
	int d = cols; // sol_dim

	for (i = 0; i < n; i++) {
		if (sign(a[r[i]][c[i]]) == 0) {
			for(j = i + 1; j < rows; j++) {
				if (sign(a[r[j]][c[i]]) == 0)
					continue;
				swap(r[i], r[j]);
				break;
			};
			if (j == rows) {
				if (i + col_counter >= cols - 1) // we don't have to swap anymore
					break;
				col_counter++;
				swap(c[i], c[cols - col_counter]);
				i--;
				continue;
			};
		};

		d--;
		if(d == 0)
			return 0;

		for(j = i + 1; j < cols; j++)
			a[r[i]][c[j]] /= a[r[i]][c[i]];
		a[r[i]][c[i]] = 1;

		for (j = i + 1; j < rows; j++) { // elim col i
			if(a[r[j]][c[i]] != 0) {
				for (k = i + 1; k < cols; k++)
					a[r[j]][c[k]] -= a[r[i]][c[k]] * a[r[j]][c[i]];
				a[r[j]][c[i]] = 0;
			};
		};
	};
	// matrix has now 'triangular' shape
	for(i = cols - d - 2; i >= 0; i--) {   // which row to eleminate
		for(j = i + 1; j < cols - d; j++) { // col to handle
			for(k = cols - d; k < cols; k++)
				a[r[i]][c[k]] -= a[r[i]][c[j]] * a[r[j]][c[k]];
			a[r[i]][c[j]] = 0;
		};
	};
	// matrix has now 'diagonal' shape
	return d;
};

/*******************************************************************
friends
*******************************************************************/

ostream& operator << (ostream &s, matrix &m) {
	int i, j, k;
	int *max_len;
	max_len = new int[m.cols];
	// determine max len of each col
	for (i = 0; i < m.cols; i++) {
		max_len[i] = 0;
		for (j = 0; j < m.rows; j++) {
			k = len(m.a[j][i]);
			if(max_len[i] < k)
				max_len[i] = k;
		};
	};
	for (i = 0; i < m.rows; i++) {
		s << "[ ";
		for (j = 0; j < m.cols; j++) {
			for(k = 0; k < (max_len[j]  - len(m.a[i][j]) + 1) / 2; k++)
				s << " ";
			s << m.a[i][j];
			for(k = 0; k < (max_len[j]  - len(m.a[i][j])) / 2; k++)
				s << " ";
			if(j != m.cols - 1)
				s << "  ";
		};
		s << " ]\n";
		if( i != m.rows - 1) {
			s << "[ ";
			for (j = 0; j < m.cols; j++) {
				for(k = 0; k < max_len[j]; k++)
					s << " ";
				if(j != m.cols - 1)
					s << "  ";
			};
			s << " ]\n";
		};
	};
	delete[] max_len;
	return s;
};

istream& operator >> (istream &s, matrix &m) {
	int i, j;
	for (i = 0; i < m.rows; i++)
		for (j = 0; j < m.cols; j++)
			s >> m.a[i][j];
	return s;
};

/*******************************************************************
ideal
*******************************************************************/

struct ideal {
	// data
	int no_of_gens;
	polynomial *gen;
	// constructors
	ideal(int);
	// destructor
	~ideal();
	// elements
// int is_element_of_Lt(monom m);
	// friends
	friend ostream& operator << (ostream &, ideal &X);
	friend istream& operator >> (istream &, ideal &X);
};

ideal::ideal(int s){
	no_of_gens = s;
	gen = new polynomial[s];
};

ideal::~ideal(){
	delete[] gen;
};

ostream& operator << (ostream &s, ideal &I) {
	for(int i = 0; i < I.no_of_gens; i++)
		s << I.gen[i] << "\n";
	return s;
};

/*******************************************************************
scheme
*******************************************************************/

struct scheme {
	// data
	int         s;                   // number of points
	matrix      *P;                  // points
	matrix      *ev_q;
	polynomial  *q;
	int         *pi;
	ideal       *separator;          // separators (watch the order)
	ideal       *Ix;                 // reduced Groebner basis
	ideal       *Lf_Ix;
	ideal       *min_Ix;             // minimal Macaulay basis
	ideal       *socle;              // socle
	ideal       *Can_Lf;             // can form of B's w.r.t. Lf(I)
	monom       *O;                  // order ideal
	monom       *B;                  // border basis
	int         no_of_border;
	int         *hilb_fct;           // Hilbert function
	int         *delta_fct;          // first difference function
	int         sigma;
	int         delta;
	int         alpha;
	int         is_CB;
	int         is_ar_Go;
	matrix      *beta_matrix;
	int         *basis;
	int         no_of_pr;
	int         *pr_deg;
	int         *pr;
	int         A_d_cols;
	int         A_d_maxdeg;
	int         *A_d_coldeg;
	polynomial  *A_d;
	int         sep;
	int         linid;
	int         linom;
	int         *(*linres);
	// constructors
	scheme(int ss);
	// destructor
	~scheme();
	// elements

	// friends
	friend istream& operator >> (istream &s, scheme &X);
	friend ostream& operator << (ostream &s, scheme &X);
};

/*******************************************************************
constructors
*******************************************************************/

scheme::scheme(int ss) {
	s              = ss;
	P              = new matrix(dim, s);
	separator      = new ideal(s);
	q              = new polynomial[s];
	pi             = new int[s];
	ev_q           = new matrix(s, s);
	O              = new monom[s];
	B              = NULL;
	Ix             = NULL;
	Lf_Ix          = NULL;
	socle          = NULL;
	Can_Lf         = NULL;
	min_Ix         = NULL;
	beta_matrix    = NULL;
	basis          = NULL;
	hilb_fct       = new int[s];
	delta_fct      = new int[s];
	no_of_border   = 0;
	delta          = 0;
	sigma          = 0;
	alpha          = 0;
	is_CB          = 0;
	is_ar_Go       = 0;
	no_of_pr       = 0;
	pr_deg         = NULL;
	pr             = NULL;
	A_d_cols       = 0;
	A_d_maxdeg     = 0;
	A_d_coldeg     = NULL;
	A_d            = NULL;
	sep            = 0;
	linid          = 0;
	linom          = 0;
	linres         = new int*[2];
	linres[0]      = new int[dim];
	linres[1]      = new int[dim];
};

/*******************************************************************
destructor
*******************************************************************/

scheme::~scheme() {
	delete   P;
	delete   separator;
	delete[] O;
	delete[] B;
	delete   min_Ix;
	delete   Can_Lf;
	delete   Ix;
	delete   Lf_Ix;
	delete   socle;
	delete[] delta_fct;
	delete[] hilb_fct;
	delete   beta_matrix;
	delete[] basis;
	delete[] pr_deg;
	delete[] pr;
	delete[] A_d_coldeg;
	delete[] A_d;
	if(sep == 0) {
		delete[] pi;
		delete[] q;
		delete   ev_q;
	};
	delete[] linres[0];
	delete[] linres[1];
	delete[] linres;
};

/*******************************************************************
I/O
*******************************************************************/

istream& operator >> (istream &s, scheme &X) {
	int i, j;
	for (i = 0; i < X.s; i++)
		for(j = 0; j < dim; j++)
			s >> X.P->a[i][j];
	return s;
};

void disp_scheme(ostream &s, scheme &X) {
	int i, j;
	for(i = 0; i < X.s; i++) {
		s << "<1";
		for(j = 0; j < dim; j++)
			s << ":" << X.P->a[i][j];
		s << ">";
	};
};

void order_ideal(ostream &s, scheme&);
void hilb_fct(ostream &s, scheme&);
void delta_fct(ostream &s, scheme&);
void sigma_X(ostream &s, scheme &);
void delta_X(ostream &s, scheme &);
void alpha_X(ostream &s, scheme &);
void is_CB(ostream &s, scheme &);
void is_ar_Go(ostream &s, scheme &);
void deg_min(ostream &s, scheme&);
void disp_socle(ostream &s, scheme&);
void disp_omega(ostream &s, scheme &);
void disp_linid(ostream &s, scheme &);
void disp_linom(ostream &s, scheme &);
void disp_linres(ostream &s, scheme &);

ostream& operator << (ostream &s, scheme &X) {
	s << "scheme consists of " << X.s << " points:\n";
	disp_scheme(s, X);
	s << "\n" << Line;
	if (X.delta != 0) {
		order_ideal(s, X);
		hilb_fct(s, X);
		delta_fct(s, X);
		sigma_X(s, X);
		delta_X(s, X);
		alpha_X(s, X);
	};
	if(X.sep == 1) {
		is_CB(s, X);
		is_ar_Go(s, X);
	};
	if (X.min_Ix!= NULL)
		deg_min(s, X);
	if (X.socle != NULL)
		disp_socle(s, X);
	if (X.no_of_pr != 0)
		disp_omega(s, X);
	if (X.linid != 0 && X.linom != 0) {
		disp_linres(s, X);
		return s;
	};
	if (X.linom != 0)
		disp_linom(s, X);
	if (X.linid != 0)
		disp_linid(s, X);
	return s;
};

/*******************************************************************
Buchberger - Mller - algorithm : Bu_Mo
*******************************************************************/

int is_el_OO(scheme &X, monom &m) {
	for(int i = 0; i < X.s; i++)
		if(X.O[i] == m)
			return i;
	return -1;
};

int element(list<mmm> &L, mmm &x) {
	part_of_list<mmm> *ptr;
	if (L.no_of_items == 0)
		return 0;
	ptr = L.first;
	while(ptr != L.last) {
		if (ptr->item == x) {
			ptr->item.ins++;
			return 1;
		};
		ptr = ptr->next;
	};
	if(ptr->item == x) {
		x.ins++;
		return 1;
	};
	return 0;
};

void hilbert(scheme&);
void hilbert2(scheme&);

void bu_mo(scheme &X) {
	int s;
	s = X.s;
// groebner basis
	list<polynomial> G;
// L = working list
	list<mmm> L;
	list<monom> B;
// working var's
	polynomial p;
	mmm m;
	K tmp;
	monom t;
// table of values
	K *ev_p, *ev_p_old, *ptr, *ev_O;
	ev_O = new K[s * s];
	ev_p = new K[s];
	ev_p_old = new K[s];

//prepare list
	m.t = 0;
	m.n = -1;
	m.k = -1;
	m.ins = 0;
	L.add(m);
	int i, j, r = -1;
	if(verbose) cerr << "calculating triangular sequence ...\n";
//start
	while(L.no_of_items) {  // list not empty
		L.min_del(m);
		p = t = m.t;

//    if (!is_element_of_Lt(G, t))
		if (m.ins == t.vars()) {
			if (m.n < 0) {
				for(i = 0; i < s; i++) {
					ev_p[i] = 1;
					ev_O[i] = 1;
				};
			}
			else {
				for(i = 0; i < s; i++)
					ev_p[i] = ev_O[m.n * s + i] * X.P->a[i][m.k];
				if (r < s - 1) {
					for(i = 0; i < s; i++)
						ev_O[(r + 1) * s + i] = ev_p[i];
				};
			};
			for (i = 0; i <= r; i++) {
				p -= X.q[i] * ev_p[X.pi[i]];
				ptr = ev_p_old;
				ev_p_old = ev_p;
				ev_p = ptr;
				for (j = 0; j < s; j++)
					ev_p[j] = ev_p_old[j] - ev_p_old[X.pi[i]] * X.ev_q->a[i][j];
			};
			for (i = 0; i < s; i++) {
				if (sign(ev_p[i]) != 0) {
					tmp = ev_p[i];
					break;
				};
			};
			if (i == s) { // p(P[i]) = 0 for all i
				B.add(t);
				G.add(p);
			}
			else {
				r++;
				X.pi[r] = i;
				tmp.invert();
				X.q[r] = p *= tmp;
				for(j = 0; j < i; j++)
					X.ev_q->a[r][j] = 0;
				X.ev_q->a[r][i] = 1;
				for(j = i + 1; j < s; j++)
					X.ev_q->a[r][j] = ev_p[j] * tmp;
				X.O[r] = t; // t -> O
				for(i = 0; i < dim; i++) {
					m.t.exp[i]++;
					m.n = r;
					m.k = i;
					if(!element(L, m)){
						m.ins = 1;
						L.add(m);
					};
					m.t.exp[i]--;
				};
				if(verbose) cerr << "[" << s - r << "]";
			};
		}
		else
			B.add(t);
	};

// create Ix
	X.Ix = new ideal(G.no_of_items);
	for(i = 0; i < X.Ix->no_of_gens; i++)
		G.min_del(X.Ix->gen[i]);

// create B
	X.B = new monom[B.no_of_items];
	X.no_of_border = B.no_of_items;
	for(i = 0; i < X.no_of_border; i++)
		B.min_del(X.B[i]);
	if(verbose) cerr << "\n";
	cout << Line;
	hilbert(X);
};

void sep(scheme &X) {
	if(verbose) cerr << "calculating biorthogonal sequence ...\n";
	int i, h;
	for(i = X.s - 1; i >= 0; i--) {
		X.separator->gen[X.pi[i]] = X.q[i];
		for(h = i + 1; h < X.s; h ++)
			X.separator->gen[X.pi[i]] -= X.separator->gen[X.pi[h]] * X.ev_q->a[i][X.pi[h]];
		if(verbose) cerr << "[" << i + 1 << "]";
	};
	if(verbose) cerr << "\n";
	cout << Line;
	X.sep = 1;
// clean up
	delete[] X.q;
	delete X.ev_q;
	delete[] X.pi;
	hilbert2(X);
};

/*******************************************************************
hilbert
*******************************************************************/

void hilbert(scheme &X) {
	int i;
	int tmp = 0;

	for (i = 0; i < X.s; i++)
		X.delta_fct[i] = 0;
	for (i = 0; i < X.s; i++)
		X.delta_fct[X.O[i].tot_deg()]++;
	for (i = 0; X.delta_fct[i] != 0 && i < X.s; i++)
	X.sigma = i - 1;
	X.delta = X.delta_fct[i - 1];
	X.alpha = X.Ix->gen[0].term[0].tot_deg();
	for (i = 0; i < X.s; i++) {
		tmp += X.delta_fct[i];
		X.hilb_fct[i] = tmp;
	};
};

void hilbert2(scheme &X) {
	int i;
	X.is_CB = 1;
	for(i = 1; i < X.s; i++) {
		if (X.separator->gen[i].term[0].tot_deg() !=
			 X.separator->gen[i - 1].term[0].tot_deg()) {
			X.is_CB = 0;
			break;
		};
	};
	if(X.is_CB) {
		X.is_ar_Go = 1;
		for(i = 0; i < X.sigma / 2 + 1; i++) {
			if(X.hilb_fct[i] + X.hilb_fct[X.sigma - i] != X.s) {
				X.is_ar_Go = 0;
				break;
			};
		};
	};
};

void alpha_X(ostream &s, scheme &X) {
	if (X.delta == 0)
		bu_mo(X);
	s << "alpha = " << X.alpha << "\n" << Line;
};

void sigma_X(ostream &s, scheme &X) {
	if (X.delta == 0)
		bu_mo(X);
	s << "sigma = " << X.sigma << "\n" << Line;
};

void delta_X(ostream &s, scheme &X) {
	if (X.delta == 0)
		bu_mo(X);
	s << "delta = " << X.delta << "\n" << Line;
};

void delta_fct(ostream &s, scheme &X) {
	if (X.delta == 0)
		bu_mo(X);
	s << "delta H(n) = ";
	for (int i = 0; i < X.sigma + 2; i++) {
		s << X.delta_fct[i] << " ";
	};
	s << "\n" << Line;
};

void hilb_fct(ostream &s, scheme &X) {
	if (X.delta == 0)
		bu_mo(X);
	s << "H(n) = ";
	for (int i = 0; i < X.sigma + 2; i++) {
		s << X.hilb_fct[i] << " ";
	};
	s << "\n" << Line;
};

void is_CB(ostream &s, scheme &X) {
	if(X.delta == 0)
		bu_mo(X);

	if(X.is_CB) {
		s << "scheme has Cayley-Bacharach property\n";
		s << Line;
		return;
	};

	if(X.sep == 0)
		sep(X);

	if(X.is_CB)
		s << "scheme has Cayley-Bacharach property\n";
	else
		s << "scheme has NOT Cayley-Bacharach property\n";
	s << Line;
};

void socle(scheme &);

void is_ar_Go(ostream &s, scheme &X) {
	if(X.delta == 0)
		bu_mo(X);

	if(X.sep == 0) {
		socle(X);
		X.is_ar_Go = 0;
		if(X.socle->no_of_gens == 1) {
			X.is_ar_Go = 1;
			X.is_CB = 1;
		};
	};

	if(X.is_ar_Go)
		s << "scheme is arithmetically Gorenstein\n";
	else
		s << "scheme is NOT arithmetically Gorenstein\n";
	s << Line;
};

/*******************************************************************
groebner

input  : zero-dimensional reduced scheme X

output : groebner basis of Ix
*******************************************************************/

void groebner(scheme &X) {
	if (X.delta == 0)
		bu_mo(X);
	cout << "Groebner basis of I(X) : \n";
	for(int i = 0; i < X.Ix->no_of_gens; i++) {
		counter = 1 + len(i + 1) + 3;
		cout << "g" << i + 1 << " = " << X.Ix->gen[i] << "\n";
	};
	cout << Line;
};

void order_ideal(ostream &s, scheme &X) {
	if (X.delta == 0)
		bu_mo(X);
	s << "order ideal of X : \n";
	for (int i = 0; i < X.s; i++) {
		s << X.O[i];
		if (i != X.s - 1)
			s << ", ";
	};
	s << "\n" << Line;
};

void separators(scheme &X) {
	if (X.delta == 0)
		bu_mo(X);

	if (X.sep == 0)
		sep(X);

	cout << "separators of X : \n";
	for (int i = 0; i < X.s; i++) {
		counter = 1 + len(i + 1) + 3;
		cout << "f" << i + 1 << " = " << X.separator->gen[i] << "\n";
	};
	cout << Line;
};

/*******************************************************************
beta_matrix
*******************************************************************/

void disp_beta_matrix(scheme &X);

void beta_matrix(scheme &X) {
	int i, j, k;
	int b = -1;       // number of basis-elements found  - 1
	int flag;
	K tmp;
	// prepare scheme
	if (X.delta == 0)
		bu_mo(X);

	if (X.sep == 0)
		sep(X);

	// prepare basis
	X.basis = new int[X.delta];

	// prepare matrix
	X.beta_matrix = new matrix(X.delta, X.s);
	X.beta_matrix->init();

	matrix basis_matrix(X.delta, X.delta);

	// prepare Lf
	polynomial *red_Lf;
	red_Lf = new polynomial[X.s];
	for(i = 0; i < X.s; i++)
		red_Lf[i] = Lf(X.separator->gen[i]);

	// start
	if(verbose) cerr << "calculating beta matrix ...\n";
	for(i = X.s - 1; i >= X.s - X.delta; i--) {
		if(verbose) cerr << "[" << X.delta - (X.s - i) + 1 << "]";
		flag = -1;
		for(j = 0; j < X.s; j++) {
			if (red_Lf[j].term[0] == X.O[i]) {
				if (flag < 0) {      // basis element
					b++;
					X.basis[b] = j;
					flag = j;
					for(k = 0; k < b; k++)
						basis_matrix.a[b][k] = - X.beta_matrix->a[j][k];
					basis_matrix.a[b][b] = 1;
					for(k = 0; k < b; k++)
						X.beta_matrix->a[j][k] = 0;
					X.beta_matrix->a[j][b] = 1;
				}
				else {               // reduce
					tmp = red_Lf[j].coef[0] / red_Lf[flag].coef[0];
					red_Lf[j] -= red_Lf[flag] * tmp;
					for(k = 0; k <= b; k++)
						X.beta_matrix->a[j][k] += tmp * basis_matrix.a[b][k];
				};
			};
		};
	};
	if(verbose) cerr << "\n";
	// clean up
	delete[] red_Lf;
};

void disp_beta_matrix(scheme &X) {
	if(X.beta_matrix == 0)
		beta_matrix(X);

	cout << "the beta_matrix with respect to the following separators :\n";
	for(int i = 0; i < X.delta; i++)
		cout << "f" << X.basis[i] + 1 << ", ";
	cout << "\n" << *X.beta_matrix << Line;
};

/*******************************************************************

*******************************************************************/

int is_el_G(scheme &X, monom &m) {
	for(int i = 0; i < X.Ix->no_of_gens; i++) {
		if(X.Ix->gen[i].term[0] == m)
			return i;
		if(X.Ix->gen[i].term[0] > m)
			return -1;
	};
	return -1;
};

int is_el_B(scheme &X, monom &m) {
	for(int i = 0; i < X.no_of_border; i++) {
		if(X.B[i] == m)
			return i;
		if(X.B[i] > m)
			return -1;
	};
	return -1;
};

int is_el_O(scheme &X, monom &m) {
	int deg = m.tot_deg();
	if (deg == 0)
		return 0;
	if (deg > X.sigma + 1)
		return -1;
	for(int i = X.hilb_fct[deg - 1]; i < X.hilb_fct[deg]; i++)
		if(X.O[i] == m)
			return i;
	return -1;
};

void create_Can_Lf(scheme &X) {
	int i, j, k, l;
	polynomial tmp, tmp1;
	monom t, tt;

	X.Can_Lf = new ideal(X.no_of_border);
	for(i = 0; i < X.no_of_border; i++) {
		t = X.B[i];
		l = is_el_G(X, t);
		if(l >= 0) {
			if (X.Lf_Ix->gen[l].no_of_monoms == 1)
				X.Can_Lf->gen[i] = 0;
			else {
				X.Can_Lf->gen[i] = t;
				X.Can_Lf->gen[i] -= X.Lf_Ix->gen[l];
			};
		}
		else {
			X.Can_Lf->gen[i] = 0;
			for(j = i - 1; j >= 0; j--) {
				if(X.B[j].tot_deg() == t.tot_deg() - 1) {
					if(Zn_order_le(X.B[j], t)) { // m found
						tt = t / X.B[j]; // tt = Xi
						tmp = X.Can_Lf->gen[j];
						tmp *= tt;
						for(k = 0; k < tmp.no_of_monoms; k++) {
							if(is_el_O(X, tmp.term[k]) >= 0) {
								tmp1 = tmp.term[k];
								tmp1 *= tmp.coef[k];
								X.Can_Lf->gen[i] += tmp1;
							}
							else
								X.Can_Lf->gen[i] += X.Can_Lf->gen[is_el_B(X, tmp.term[k])]
													  * tmp.coef[k];
						}; // end k
						break;
					}; // end if
				}; // end if
			}; // end j
		}; // end else
	};
};

void create_Lf_Ix(scheme &X) {
	int i;
	X.Lf_Ix = new ideal(X.Ix->no_of_gens);
	for(i = 0; i < X.Ix->no_of_gens; i++)
		X.Lf_Ix->gen[i] = Lf(X.Ix->gen[i]);
};

void min_Ix(scheme &X) {
	// calculate a minimal Macaulay basis for Ix
	// we need degree compatible term ordering

	int i, j, k, kk, l, d;
	polynomial *b;
	monom *M;
	polynomial tmp, tmp1;
	monom t, tt, L, ti, tj;
	int *alpha;
	int *beta;
	int *c;
	int *r;

	int no = 0;

	if (X.delta == 0)
		bu_mo(X);

	if (X.Lf_Ix == NULL)
		create_Lf_Ix(X);

	if (X.Can_Lf == NULL)
		create_Can_Lf(X);

	// step 1
	b = new polynomial[X.no_of_border];
	matrix RV(X.Ix->no_of_gens, X.no_of_border);
	RV.init();

	for(i = 0; i < X.no_of_border; i++) {
		t = X.B[i];
		l = is_el_G(X, t);
		if(l >= 0) {
			b[i] = X.Lf_Ix->gen[l];
			RV.a[i][l] = 1;
		}
		else {
			for(j = i - 1; j >= 0; j--) {
				if(X.B[j].tot_deg() == t.tot_deg() - 1) {
					if(Zn_order_le(X.B[j], t)) { // m found
						tt = t / X.B[j]; // tt = Xi
						tmp = X.Can_Lf->gen[j];
						tmp *= tt;
						for(k = 0; k < tmp.no_of_monoms; k++) {
							if(is_el_O(X, tmp.term[k]) < 0) {
								l = is_el_B(X, tmp.term[k]);
								for(kk = 0; kk < RV.cols; kk++)
									RV.a[i][kk] += RV.a[l][kk] * tmp.coef[k];
							};
						};
						break;
					};
				};
			};
			b[i] = t;
			b[i] -= X.Can_Lf->gen[i];
		};
	};

	// step 2
	alpha = new int[(X.Ix->no_of_gens - 1) * X.Ix->no_of_gens / 2];
	beta  = new int[(X.Ix->no_of_gens - 1) * X.Ix->no_of_gens / 2];
	M = new monom[(X.Ix->no_of_gens - 1) * X.Ix->no_of_gens / 2];

	d = X.O[0].dim;

	for(i = 0; i < X.Ix->no_of_gens - 1; i++) {
		for(j = i + 1; j < X.Ix->no_of_gens; j++) {
			L = lcm(X.Ix->gen[i].term[0], X.Ix->gen[j].term[0]);
			if(L.tot_deg() - 2 > X.sigma + 1)
				continue;
			alpha[no] = -1;
			beta[no]  = -1;
			for(k = 0; k < d; k++) {
				kk = X.Ix->gen[i].term[0].exp[k] - X.Ix->gen[j].term[0].exp[k];
				if(kk > 0)  // Xalpha found
					alpha[no] = k;
				else if(kk < 0) // Xbeta found
					beta[no]  = k;
			};
			t = 0;
			t.exp[alpha[no]]++;
			t.exp[beta[no]]++;
			M[no] = L / t;
			if(is_el_O(X, M[no]) >= 0)
				no++;
		}; // end j
	}; // end i

	matrix R(RV.cols, no);
	R.init();

	// step 3

	for(i = 0; i < no; i++) {
		ti = 0;
		ti.exp[alpha[i]]++;
		tj = 0;
		tj.exp[beta[i]]++;

		t = M[i] * ti; // == Xalpha * M
		tmp = b[is_el_B(X, t)];
		tmp *= tj;

		t = M[i] * tj; // == Xbeta * M
		tmp1 = b[is_el_B(X, t)];
		tmp1 *= ti;
		tmp -= tmp1;

		for(j = 0; j < tmp.no_of_monoms; j++) {
			l = is_el_B(X, tmp.term[j]);
			if(l >= 0)
				for(k = 0; k < R.cols; k++)
					R.a[i][k] -= tmp.coef[j] * RV.a[l][k];
		};
	};

	c = new int[X.Ix->no_of_gens];
	for(i = 0; i < R.cols; i++)
		c[i] = i;
	r = new int[no];
	for(i = 0; i < R.rows; i++)
		r[i] = i;

	d = R.solve(c, r);

	X.min_Ix = new ideal(d);

	int deg, next = 0;

	for(i = 0; i < d; i++) {
		deg = X.sigma + 3;
		for(j = X.Ix->no_of_gens - d; j < X.Ix->no_of_gens - i; j++) {
			if (X.Ix->gen[c[j]].term[0].tot_deg() < deg) {
				deg = X.Ix->gen[c[j]].term[0].tot_deg();
				next = j;
			};
		};
		hom(X.min_Ix->gen[i], X.Ix->gen[c[next]]);
		c[next] = c[X.Ix->no_of_gens - i - 1];
	};
	delete[] c;
	delete[] r;
	delete[] b;
	delete[] M;
	delete[] alpha;
	delete[] beta;
};

void disp_min(scheme &X) {
	if (X.min_Ix == NULL)
		min_Ix(X);

	cout << "a minimal basis for the projective vanishing ideal consists of the following\n"
		  << X.min_Ix->no_of_gens << " polynomials :\n";

	for(int i = 0; i < X.min_Ix->no_of_gens; i++) {
		counter = 3 + len(i + 1) + 3;
		cout << "min" << i + 1 << " = " << X.min_Ix->gen[i] << "\n";
	};
	cout << Line;
};

void deg_min(ostream &s, scheme &X) {
	if (X.min_Ix == NULL)
		min_Ix(X);
	s << "the degrees of the minimal generators of the projective vanishing ideal are:\n";
	int d = X.min_Ix->gen[0].term[0].tot_deg();
	int n = 0;
	for(int i = 0; i < X.min_Ix->no_of_gens; i++) {
		if(X.min_Ix->gen[i].term[0].tot_deg() != d) {
			s << "deg " << d << " : " << n << " generator(s)\n";
			n = 1;
			d = X.min_Ix->gen[i].term[0].tot_deg();
		}
		else n++;
		if(i == X.min_Ix->no_of_gens - 1)
			s << "deg " << d << " : " << n << " generator(s)\n";
	};
	s << Line;
};

/*******************************************************************
is_delta_uf
*******************************************************************/

void is_delta_uf(scheme &X) {
	if(X.delta == 0)
		bu_mo(X);

	if(X.sep == 0)
		sep(X);

	if(X.delta == 1) {
		if(X.is_CB) {
			cout << "scheme has Cayley-Bacharach property\n";
			cout << "since delta = 1, scheme has also delta uniform property\n";
			cout << "and is even 2-uniform\n";
		}
		else {
			cout << "scheme has NOT Cayley-Bacharach property\n";
			cout << "hence it cannot have delta uniform property\n";
		};
		cout << Line;
	}
	else {   // we have to look at beta_matrix
		if(X.beta_matrix == 0)
			beta_matrix(X);

		// prepare
		int i;
		int *c, *r;
		c = new int[X.delta];
		r = new int[X.delta + 1];
		for(i = 0; i < X.delta; i++) {
			c[i] = i;
			r[i] = i;
		};
		r[X.delta] = X.s;
		int quit = 0;

		// calc minors
		while (!quit) {
			if(sign(X.beta_matrix->minor(X.delta, c, r)) == 0) {
				cout << "\nlinear dependence between ";
				for(i = 0; i < X.delta; i++)
					cout << "f" << r[i] + 1 << ", ";
				cout << "\ntherefore ";
				break;
			};
			for(i = 0; i < X.delta; i++) {
				r[i]++;
				if(i == X.delta - 1)
					if(verbose) cout << "[" << X.s - r[i] + 1 << "]";
				if (r[i] == r[i + 1]) {
					r[i] = i;
					if(i == X.delta - 1) {
						quit = 1;
						break;
					};
				}
				else break;
			};
		};
		if (verbose) cout << "\n";
		if(quit) // is delta uniform
			cout << "scheme has delta uniform property\n";
		else
			cout << "scheme has NOT delta uniform property\n";
		cout << Line;
		delete[] c;
		delete[] r;
	};
};

/*******************************************************************
is_n_uf
*******************************************************************/

void is_n_uf(scheme &X) {
	int n;
	cin >> n;

	if(n <= 0) {
		cout << "n-unifomity only defined for n > 0 !!!\n";
		cout << Line;
		return;
	};

	if(X.delta == 0)
		bu_mo(X);

	if(n == X.delta) {
		is_delta_uf(X);
		return;
	};

	if(n > X.delta) {
		cout << "sorry, but checking n-unifomity for n > delta is not supported yet\n";
		cout << Line;
		return;
	};

	if(X.beta_matrix == 0)
		beta_matrix(X);

	// prepare
	int i, j;
	int *r = new int[n + 1];
	int *cc = new int[X.delta];
	int *rr = new int[X.delta];
	for(i = 0; i < n; i++)
		r[i] = i;
	r[n] = X.s;
	int quit = 0;

	matrix m(X.delta, n);
	// calc minors
	while (!quit) {
		for(i = 0; i < n; i++) {
			rr[i] = i;
			for(j = 0; j < X.delta; j++) {
				cc[j] = j;
				m.a[i][j] = X.beta_matrix->a[r[i]][j];
			};
		};
		j = m.solve(cc, rr);
		if(j > X.delta - n) {
			cout << "\nlinear dependence between ";
			for(i = 0; i < n; i++)
				cout << "f" << r[i] + 1 << ", ";
			cout << "\ntherefore ";
			break;
		};
		for(i = 0; i < n; i++) {
			r[i]++;
			if(i == n-1)
				if(verbose) cout << "[" << X.s - r[i] + 1 << "]";
			if (r[i] == r[i + 1]) {
				r[i] = i;
				if(i == n - 1) {
					quit = 1;
					break;
				};
			}
			else break;
		};
	};
	if(quit) // is n uniform
		cout << "\nscheme has " << n << "-uniform property\n";
	else
		cout << "scheme has NOT " << n << "-uniform property\n";
	cout << Line;
	delete[] r;
	delete[] cc;
	delete[] rr;
};

/*******************************************************************
												 _
socle       calculating the socle of R
*******************************************************************/

void socle(scheme &X) {
	int i, j, jj, k, l, ll, cols, deg, block, found;
	int socle_dim = 0;
	int *socle_d;
	int *c, *r;
	list<polynomial> s;

	monom tmp;
	polynomial pol, pol1;

	if (X.delta == 0)
		bu_mo(X);

	if (X.socle != NULL)
		return;

	if (X.Lf_Ix == NULL)
		create_Lf_Ix(X);

	if (X.Can_Lf == NULL)
		create_Can_Lf(X);

	c = new int[X.s];
	r = new int[X.delta_fct[1] * X.s];

	socle_d = new int[X.sigma + 2];

	// add elements of degree sigma + 1
	for(i = X.s - 1; i >= X.hilb_fct[X.sigma]; i--)
		s.add(polynomial(X.O[i]));

	for(deg = X.sigma; deg >= X.alpha - 1 && deg > 0; deg--) {  // degree
		for(i = X.hilb_fct[deg] - 1; i > X.hilb_fct[deg - 1] - 1; i--) { // monom
			found = 0;
			for(k = 1; k < X.hilb_fct[1] && found == 0; k++) {
				tmp = X.O[i] * X.O[k];
				if(is_el_O(X, tmp) >= 0) {
					found = 1;
					break;
				};
			};
			if(found == 0) {  // socle element of gr_sigma R
				cols = i + 1 - X.hilb_fct[deg - 1];
				matrix m(cols, X.delta_fct[1] * X.delta_fct[deg + 1]);
				m.init();
				for(j = 0; j < m.cols; j++)
					c[j] = j;
				for(j = 0; j < m.rows; j++)
					r[j] = j;
				for(j = i; j >= X.hilb_fct[deg - 1]; j--) {  // smaller monoms
					jj = i - j; // = col
					block = X.delta_fct[deg + 1];
					for(k = 1; k < X.hilb_fct[1]; k++) {
						tmp = X.O[j] * X.O[k];
						l = is_el_O(X, tmp);
						if(l >= 0)
							m.a[(k - 1) * block + (l - X.hilb_fct[deg])][jj] = 1;
						else { // found == 0;  we have to build Can_Lf
							l = is_el_B(X, tmp);
							if(sign(X.Can_Lf->gen[l]) != 0)
								for(ll = 0; ll < X.Can_Lf->gen[l].no_of_monoms; ll++)
									m.a[(k - 1) * block + (is_el_O(X, X.Can_Lf->gen[l].term[ll]) - X.hilb_fct[deg])][jj] = X.Can_Lf->gen[l].coef[ll];
						}; // end if
					}; // end k
				}; // end j : all smaller monoms treated
				socle_d[deg] = m.solve(c, r);
				if(socle_d[deg] > 0) {
					socle_dim += socle_d[deg];
				};
				for(l = 0; l < socle_d[deg]; l++) {
					pol = 0;
					for(ll = 0; ll < cols - socle_d[deg]; ll++) {
						pol1 = X.O[cols - c[ll] - 1+ X.hilb_fct[deg - 1]];
						pol1 *= m.a[r[ll]][c[cols - l - 1]];
						pol -= pol1;
					};
					pol1 = X.O[cols - c[cols - l - 1] - 1 + X.hilb_fct[deg - 1]];
					pol += pol1;
					s.add(pol);
				};
				break; // treat next degree
			}; // end if
		}; // end i
	}; // end deg

	X.socle = new ideal(s.no_of_items);
	for(i = 0; i < X.socle->no_of_gens; i++)
		s.min_del(X.socle->gen[i]);

	delete[] c;
	delete[] r;
	delete[] socle_d;

};

void disp_socle(ostream &s, scheme &X) {
	if(X.socle == NULL)
		socle(X);

	s << "the socle of R/(X_0) is generated by the following " << X.socle->no_of_gens <<
			  " polynomials :\n";

	for(int i = 0; i < X.socle->no_of_gens; i++)
		s << "soc" << i + 1 << " = " << X.socle->gen[i] << "\n";

	s << Line;
};

void typ(scheme &X) {
	if(X.socle == NULL)
		socle(X);
	cout << "type of X is " << X.socle->no_of_gens << "\n" << Line;
};

/*******************************************************************
omega       calculating a minimal presentation of the canonical module or R
*******************************************************************/

int check(polynomial &f, monom &t) {
	for(int i = 0; i < f.no_of_monoms; i++)
		if (f.term[i] == t)
			return i;
	return -1;
};

polynomial& Can_OO(polynomial &f, monom *OO, polynomial *Can, int o, scheme &X) {
	int l, l1;
	int d = dim;
	dim = proj;
	polynomial tmp;
	monom tt;
	K cc;
	for(l = 0; l < f.no_of_monoms; l++) {
		l1 = is_el_O(X, f.term[l]);
		if(l1 >= 0)
			continue;
		tmp = f.term[l];
		tmp *= f.coef[l];
		f -= tmp; // remove term
		tt = tmp.term[0];
		for(l1 = 0; l1 < o; l1++)
			if (tt == OO[l1])
				break;
		tt /= OO[l1];
		cc = tmp.coef[0];
		tmp = Can[l1];
		tmp *= tt;
		tmp *= cc;
		f += tmp; // replace with Can
		l = -1;
	};
	dim = d;
	return f;
};

void omega(scheme &X) {
	if (X.delta == 0)
		bu_mo(X);

	int s = X.s;
	int ee, g, i, j, k, l = 0, l1 = 0, m, n, o;
	int found;
	int number = 0;

	monom t, t1, tt, ttt;
	ttt.dim = proj;

	polynomial tmp;
	dim++;
	polynomial tmp1;
	dim--;

	K cc;
	// first                   _
	// determine socle of gr  (R)
	//                      sigma
	int no_gr_socle = 0;
	for(i = 1; i < s; i++) {
		found = 0;
		for(k = 1; k < X.hilb_fct[1]; k++) {
			t = X.O[i] * X.O[k];
			if(is_el_O(X, t) >= 0) {
				found = 1;
				break;
			};
		};
		if(found == 0) // socle element of gr_sigma R
			no_gr_socle++;
	};
	int *gr_socle = new int[no_gr_socle];
	int *elim = new int[s - no_gr_socle];
	int no_of_elim = 0;
	no_gr_socle = 0;
	for(i = 0; i < s; i++) {
		found = 0;
		for(k = 1; k < X.hilb_fct[1]; k++) {
			t = X.O[i] * X.O[k];
			if(is_el_O(X, t) >= 0) {
				found = 1;
				break;
			};
		};
		if(found == 0) {  // socle element of gr_sigma R
			gr_socle[no_gr_socle] = i;
			no_gr_socle++;
		}
		else {
			elim[no_of_elim] = i;
			no_of_elim++;
		};
	}; // socle determined, smaller first

	number = no_gr_socle * (no_gr_socle - 1) / 2 + no_gr_socle * dim;
	int *f = new int[number];
	monom *TT = new monom[number];
	m = 0;
	// bi-syzygies --- calc Lt's
	for(i = 0; i < no_gr_socle; i++) {
		for(j = i + 1; j < no_gr_socle; j++) {
			TT[m] = X.O[gr_socle[i]] / gcd(X.O[gr_socle[i]], X.O[gr_socle[j]]);
			f[m] = i;
			m++;
		};
	};
	// stand alone syzygy
	for(i = 0; i < no_gr_socle; i++) { // projection
		for(j = 0; j < dim; j++) { // Xj
			TT[m] = 0;
			TT[m].exp[j] = X.O[gr_socle[i]].exp[j] + 1;
			f[m] = i;
			m++;
		};
	};
	m = number;
	// bi-syzygies --- calc redundant Lt's
	for(i = 0; i < m; i++) {
		for(k = 0; k < m; k++) {
			if(k == i) continue;
			if(f[k] == -1) continue;
			if(f[k] == f[i] && Zn_order_le(TT[k], TT[i])) {
				number--;
				f[i] = -1;
				break;
			};
		};
	};

	list<monom> oo;

	for(i = 0; i < s; i++)
		for(j = i; j < s; j++) {
			t = X.O[i] * X.O[j];
			if(is_el_O(X, t) < 0)
				oo.add_set(t);
		};
	for(i = 0; i < m; i++) {
		if(f[i] == -1) continue;
		if(is_el_O(X, TT[i]) < 0) {
			for(j = 0; j < s; j++) {
				t = TT[i] * X.O[j];
				oo.add_set(t);
			};
		};
	};

	delete[] TT;

	o = oo.no_of_items;
	dim++;
	polynomial *Can = new polynomial[o];
	dim--;
	monom *OO = new monom[o];

	if(verbose) cerr << "calculating canonical forms ...\n";
	// create Can's
	for(i = 0; i < o; i++) {
		oo.min_del(OO[i]);
		if(verbose) cerr << "[" << o - i << "]";
		l = is_el_G(X, OO[i]);
		if (l >= 0) {
			if(X.Ix->gen[l].no_of_monoms == 1)
				continue;
			tmp = OO[i];
			tmp *= X.Ix->gen[l].coef[0];
			tmp -= X.Ix->gen[l];
			hom(Can[i], tmp);
			k = OO[i].tot_deg() - tmp.term[0].tot_deg();
			if (k > 0)
				for(l = 0; l < Can[i].no_of_monoms; l++)
					Can[i].term[l].exp[dim] += (char)k;
			continue;
		};
		for(l = i - 1; l >= 0; l--)
			if (Zn_order_lt(OO[l], OO[i]))
				break;
		if(sign(Can[l]) == 0)
			continue;
		t = OO[i] / OO[l];
		dim++;
		Can[i] = Can[l];
		dim--;
		Can[i] *= t;
		found = 1;
		while (found == 1) {
			found = 0;
			for(j = 0; j < Can[i].no_of_monoms; j++) {
				l = is_el_O(X, Can[i].term[j]);
				if (l < 0) {
					found = 1;
					break;
				};
			};
			if (found == 1) {
				tt = Can[i].term[j];
				for(k = i - 1; k >= 0; k--) {
					if(tt == OO[k]) {
						dim++;
						tmp1 = tt;
						cc = Can[i].coef[j];
						tmp1 *= cc;
						Can[i] -= tmp1;   // remove term
						if(sign(Can[k]) != 0) {
							tmp1 = Can[k];
							tmp1 *= cc;
							tt /= OO[k];
							tmp1 *= tt;
							Can[i] += tmp1;   // replace with Can
						};
						dim--;
						break;
					};
				};
			}; // end found == 1
		}; // end while
	}; // for i ...   => Can's created
	if(verbose) cerr << "\n" << Line;

	dim++;
	polynomial *SS = new polynomial[no_of_elim * s];
	dim--;

	// bi-syzygies for SS
	for(i = 0; i < no_of_elim; i++) {
		for(j = 0; j < no_gr_socle; j++) {
			if(! Zn_order_lt(X.O[elim[i]], X.O[gr_socle[j]])) continue;
			l = gr_socle[j];
			tt = X.O[l] / X.O[elim[i]];
			for(k = 1; k < elim[i]; k++) {
				t1 = tt * X.O[k];
				l1 = is_el_O(X, t1);
				if(l1 >= 0)
					continue;
				for(l1 = 0; l1 < o; l1++)
					if (t1 == OO[l1])
						break;
				found = check(Can[l1], X.O[l]);
				if (found >= 0) {
					SS[i * s + k].coef[0] = Can[l1].coef[found];
					SS[i * s + k].term[0].exp[dim] = Can[l1].term[found].exp[dim];
				};
			};
			for(k = elim[i] + 1; k < s; k++) {
				t1 = tt * X.O[k];
				l1 = is_el_O(X, t1);
				if(l1 >= 0)
					continue;
				for(l1 = 0; l1 < o; l1++)
					if (t1 == OO[l1])
						break;
				found = check(Can[l1], X.O[l]);
				if (found >= 0) {
					SS[i * s + k].coef[0] = Can[l1].coef[found];
					SS[i * s + k].term[0].exp[dim] = Can[l1].term[found].exp[dim];
				};
			};
			prj(tt);
			dim++;
			tmp1 = tt;
			SS[i * s + l] -= tmp1;
			dim--;
			break;
		};
	};

	dim++;
	// rewrite SS
	for(i = no_of_elim - 2; i >= 0; i--) {
		for(j = i + 1; j < no_of_elim; j++) {
			if (sign(SS[i * s + elim[j]]) == 0) continue;
			for(k = 0; k < no_gr_socle; k++) {
				tmp1 = SS[i * s + elim[j]] * SS[j * s + gr_socle[k]];
//          Can_OO(tmp1, OO, Can, o, X);
				SS[i * s + gr_socle[k]] -= tmp1;
			};
			SS[i * s + elim[j]] = 0;
		};
	};
	dim--;

	dim++;
	polynomial *S = new polynomial[number * s];
	dim--;

	int *e = new int[number];
	m = 0;
	int m1 = 0;
	// bi-syzygies
	for(i = 0; i < no_gr_socle; i++) {
		for(j = i + 1; j < no_gr_socle; j++) {
			if(f[m1] == -1) {
				m1++;
				continue;
			};
			t = gcd(X.O[gr_socle[i]], X.O[gr_socle[j]]);
			n = is_el_O(X, X.O[gr_socle[i]] / t);
			g = is_el_O(X, t);
			if(n != 0) { // nothing would happen
				for(k = 1; k < s; k++) {
					if(k == g) continue;
					t1 = X.O[n] * X.O[k];
					l1 = is_el_O(X, t1);
					if(l1 >= 0)
						continue;
					for(l1 = 0; l1 < o; l1++)
						if (t1 == OO[l1])
							break;
					found = check(Can[l1], X.O[gr_socle[i]]);
					if (found >= 0) {
						S[m * s + k].coef[0] = Can[l1].coef[found];
						S[m * s + k].term[0].exp[dim] = Can[l1].term[found].exp[dim];
					};
				};
			};
			tt = X.O[n];
			prj(tt);
			dim++;
			tmp1 = tt;
			S[m * s + gr_socle[i]] -= tmp1;
			dim--;
			n = is_el_O(X, X.O[gr_socle[j]] / t);
			if(n != 0) { // nothing would happen
				for(k = 1; k < s; k++) {
					if(k == g) continue;
					t1 = X.O[n] * X.O[k];
					l1 = is_el_O(X, t1);
					if(l1 >= 0)
						continue;
					for(l1 = 0; l1 < o; l1++)
						if (t1 == OO[l1])
							break;
					found = check(Can[l1], X.O[gr_socle[j]]);
					if (found >= 0) {
						tt = 0;
						tt.exp[proj - 1] = Can[l1].term[found].exp[proj - 1];
						dim++;
						tmp1 = tt;
						tmp1.coef[0] = Can[l1].coef[found];
						S[m * s + k] -= tmp1;
						dim--;
					};
				};
			};
			dim++;
			tt = X.O[n];
			prj(tt);
			tmp1 = tt;
			S[m * s + gr_socle[j]] += tmp1;
			dim--;
			e[m] = i;
			m++;
			m1++;
		};
	};
	// stand alone syzygy
	for(i = 0; i < no_gr_socle; i++) { // projection
		for(j = 0; j < dim; j++) { // Xj
			if(f[m1] == -1) {
				m1++;
				continue;
			};
			t = 0;
			t.exp[j] = X.O[gr_socle[i]].exp[j] + 1;
			// X_j*p_i
			for(k = 0; k < s; k++) {
				t1 = t * X.O[k];
				l = is_el_O(X, t1);
				if (l >= 0)
					continue;
				for(l = 0; l < o; l++)
					if (t1 == OO[l])
						break;
				found = check(Can[l], X.O[gr_socle[i]]);
				if (found >= 0) {
					S[m * s + k].coef[0] = Can[l].coef[found];
					S[m * s + k].term[0].exp[dim] = Can[l].term[found].exp[dim];
				};
			};
			tt = t;
			prj(tt);
			dim++;
			tmp1 = tt;
			S[m * s + gr_socle[i]] -= tmp1;
			dim--;
			e[m] = i;
			m++;
			m1++;
		};
	};
	delete[] f;
	if(verbose) cerr << number << " syzygies among " << no_gr_socle
		  << " projections found.\n" << Line;

	dim++;
	if(verbose) cerr << "rewriting syzygies ...\n";
	// start rewrite
	for(i = 0; i < no_of_elim; i++) {
		if(verbose) cerr << "[" << s - i - no_gr_socle << "]";
		for(m = 0; m < number; m++) {
			if (sign(S[m * s + elim[i]]) == 0) continue;
			for(k = 0; k < no_gr_socle; k++) {
				tmp1 = S[m * s + elim[i]] * SS[i * s + gr_socle[k]]; // rewrite
				S[m * s + gr_socle[k]] -= tmp1;
			};
			S[m * s + elim[i]] = 0;
		};
	};
	if(verbose) cerr << "\n";

	long number1 = 0;
	for(i = 0; i < no_gr_socle; i++) {
		m = 0;
		for(j = 0; j < number; j++)
			if(e[j] == i) m++;
		number1 += m * (m - 1) / 2;
	};


	matrix T(number, number);
	T.init();
	int *c = new int[number];
	int *r = new int[number];

	for(i = 0; i < number; i++)
		c[i] = r[i] = i;

	polynomial *R = new polynomial[no_gr_socle];
	monom   *L    = new monom[no_gr_socle];

	m = 0; // maxdeg
	for(i = 0; i < number; i++) {
		g = S[i * s + gr_socle[e[i]]].term[0].tot_deg();
		if(g > m) m = g;
	};

	g = 0; // sol dim
	// bi-syzygies
	if(verbose) cerr << "eliminating syzygies ...\n";
	for(i = 0; i < number; i++) { // S
		for(j = i + 1; j < number; j++) { //
			if(e[i] != e[j])
				continue;
//       cout << "i : " << i << "  j : " << j << "\n";
			l1 = 0;
			t = lcm(S[i * s + gr_socle[e[i]]].term[0], S[j * s + gr_socle[e[j]]].term[0]);
			if(t.tot_deg() > m) {
				if(verbose) cerr << "[" << number1 << "]";
				number1--;
				continue;
			};
			for(l = 0; l < no_gr_socle; l++) {
				if(S[i * s + gr_socle[l]].term[0].exp[proj - 1] == 0) {
					R[l] = Lf(S[i * s + gr_socle[l]]);
					R[l] *= t / S[i * s + gr_socle[e[i]]].term[0];
				}
				else R[l] = 0;
				if(S[j * s + gr_socle[l]].term[0].exp[proj - 1] == 0) {
					tmp1 = Lf(S[j * s + gr_socle[l]]);
					tmp1 *= t / S[j * s + gr_socle[e[i]]].term[0];
#ifdef field_QN
					if(sign(S[j * s + gr_socle[e[i]]].coef[0].a[0]) *
						sign(S[i * s + gr_socle[e[i]]].coef[0].a[0]) < 0)
#else
					if(sign(S[j * s + gr_socle[e[i]]].coef[0]) *
						sign(S[i * s + gr_socle[e[i]]].coef[0]) < 0)
#endif
						R[l] += tmp1;
					else
						R[l] -= tmp1;
				};
			};
			if(t == S[i * s + gr_socle[e[i]]].term[0]) {
				T.a[r[g]][i] = 1;
				l1 = 1;
			};
			if(t == S[j * s + gr_socle[e[j]]].term[0]) {
				T.a[r[g]][j] = 1;
#ifdef field_QN
				if(sign(S[j * s + gr_socle[e[i]]].coef[0].a[0]) *
					sign(S[i * s + gr_socle[e[i]]].coef[0].a[0]) > 0)
#else
				if(sign(S[j * s + gr_socle[e[i]]].coef[0]) *
					sign(S[i * s + gr_socle[e[i]]].coef[0]) > 0)
#endif
					T.a[r[g]][j] = -1;
				l1 = 1;
			};
			do {
//          for(n = 0; n < no_gr_socle; n++)
//             cout << R[n] << " e" << n << "\n";
				ee = -1;
				for(k = 0; k < no_gr_socle; k++) { // determine Lt(R)
					if(sign(R[k]) == 0)
						continue;
					L[k] = R[k].term[0];
					L[k] /= X.O[gr_socle[k]];
//             cout << "Lk : " << L[k] << "\n";
					if(ee == -1)
						ee = k;
					else if(L[k] > L[ee])  // ">" makes the ordering correct
						ee = k;
				};
//          cout << "ee : " << ee << "\n";
				if(ee != -1) { // i.e. R != 0
					for(l = 0; l < number; l++) {
						if (e[l] == ee) {
							if (S[l * s + gr_socle[ee]].term[0] == R[ee].term[0]) {
//                      cout << "mit : " << l << "\n";
								cc = R[ee].coef[0] / S[l * s + gr_socle[ee]].coef[0];
								for(n = 0; n < no_gr_socle; n++) {
									if(is_prj(S[l * s + gr_socle[n]])) continue;
									tmp1 = Lf(S[l * s + gr_socle[n]]);
									tmp1 *= cc;
									R[n] -= tmp1;
								};
								T.a[r[g]][l] -= cc;
								l1 = 1;
								ee = -2;
								break;
							};
						};
					};
				};
				if(ee >= 0) { // i.e. R != 0
					for(l = 0; l < number; l++) {
						if (e[l] == ee) {
							if (Zn_order_lt(S[l * s + gr_socle[ee]].term[0], R[ee].term[0])) {
//                      cout << "mit : " << l << "\n";
								cc = R[ee].coef[0] / S[l * s + gr_socle[ee]].coef[0];
								tt = R[ee].term[0] / S[l * s + gr_socle[ee]].term[0];
								for(n = 0; n < no_gr_socle; n++) {
									if(is_prj(S[l * s + gr_socle[n]])) continue;
									tmp1 = Lf(S[l * s + gr_socle[n]]);
									tmp1 *= cc;
									tmp1 *= tt;
									R[n] -= tmp1;
								};
								ee = -2;
								break;
							};
						};
					};
				};
				if (ee >= 0) {
					cerr << "\ncannot reduce to zero --> error !?! !?!\n";
					for(n = 0; n < no_gr_socle; n++) {
						if (n == ee) cout << "> ";
						cout << R[n] << "\n";
					};
					ee = -1;
				};
			} while (ee != -1);
			if(verbose) cerr << "[" << number1 << "]";
			number1--;
			if (l1 != 0)
				g = number - T.solve(c, r);
		};
	};
	if(verbose) cerr << "\n";

	for (i = 0; i < g; i++)
		e[c[i]] = -1;

	l1 = 0;
	if(verbose) cerr << "rewriting syzygies ...\n";
	// start rewrite
	for(i = 0; i < no_gr_socle; i++) {
		if(verbose) cerr << "[" << no_gr_socle - i << "]";
		for(m = 0; m < number; m++) {
			if (e[m] == -1) continue;
			// search for units
			if (sign(S[m * s + gr_socle[i]]) != 0 &&
				 sign(S[m * s + gr_socle[i]].term[0]) == 0) {
				// unit found -> rewrite
				c[l1] = i;
				l1++;
				cc = S[m * s + gr_socle[i]].coef[0];
				cc.invert();
				for(j = 0; j < no_gr_socle; j++) {
					if(j == i) continue;
					S[m * s + gr_socle[j]] *= -cc;
				};
				for(k = 0; k < number; k++) {
					if(e[k] == -1) continue;
					if(sign(S[k * s + gr_socle[i]]) == 0) continue;
					if(k == m) continue;
					for(j = 0; j < no_gr_socle; j++) {
						if(j == i) continue;
						tmp1 = S[k * s + gr_socle[i]] * S[m * s + gr_socle[j]]; // rewrite
						S[k * s + gr_socle[j]] += tmp1;
					};
					S[k * s + gr_socle[i]] = 0;
				};
				e[m] = -1;
				break; // rewrite done
			};
		};
	};
	dim--;
	if(verbose) cerr << "\n" << Line;

	X.pr = new int[no_gr_socle - l1];
	X.no_of_pr = 0;
	for(i = 0; i < l1; i++)
		gr_socle[c[i]] = -1;

	for(i = 0; i < no_gr_socle; i++)
		if(gr_socle[i] != -1) {
			X.pr[X.no_of_pr] = gr_socle[i];
			X.no_of_pr++;
		};

	delete[] c;
	c = new int[X.sigma+3];
	for(i = 0; i < X.sigma + 3; i++) c[i] = 0;

	n = 0;
	for(i = 0; i < number; i++) {
		if(e[i] == -1) continue;
		for(j = 0; j < X.no_of_pr; j++) {
			if(sign(S[i * s + X.pr[j]]) == 0) continue;
			l = S[i * s + X.pr[j]].term[0].tot_deg() +
				 S[i * s + X.pr[j]].term[0].exp[proj - 1];
			l += X.O[X.pr[X.no_of_pr - 1]].tot_deg() - X.O[X.pr[j]].tot_deg();
			c[l]++;
			if(l > n) n = l;
			break;
		};
	};

	X.A_d_maxdeg = n;
	X.A_d_coldeg = new int[n + 1];
	for(i = 0; i <= X.A_d_maxdeg; i++)
		X.A_d_coldeg[i] = c[i];

	dim++;
	X.A_d = new polynomial[1];
	delete[] X.A_d;
	X.A_d = new polynomial[(number - g - l1) * X.no_of_pr];
	//                     cols                rows
	X.A_d_cols = number - g - l1;

	n = 0;
	for(i = 0; i <= X.A_d_maxdeg; i++) {
		for(k = 0; k < number; k++) {
			if(e[k] == -1) continue;
			for(m = 0; m < X.no_of_pr; m++) {
				if(sign(S[k * s + X.pr[m]]) == 0) continue;
				l = S[k * s + X.pr[m]].term[0].tot_deg() +
					 S[k * s + X.pr[m]].term[0].exp[proj - 1];
				l += X.O[X.pr[X.no_of_pr - 1]].tot_deg() - X.O[X.pr[m]].tot_deg();
				break;
			};
			if(l == i) {
				for(m = 0; m < X.no_of_pr; m++)
					X.A_d[m * X.A_d_cols + n] = S[k * s + X.pr[m]];
				n++;
			};
		};
	};
	dim--;

	X.pr_deg = new int[X.sigma + 2];
	for(i = 0; i < X.sigma + 2; i++)
		X.pr_deg[i] = 0;

	for(i = 0; i < X.no_of_pr; i++)
		X.pr_deg[X.O[X.pr[i]].tot_deg()]++;

	delete[] S;
	delete[] OO;
	delete[] Can;
	delete[] L;
	delete[] e;
	delete[] c;
	delete[] r;
	delete[] R;
	delete[] gr_socle;
};

void disp_omega(ostream &s, scheme &X) {
	if(X.no_of_pr == 0)
		omega(X);

	s << "the minimal homogeneous presentation of the canonical module w_R :\n";

	int l = X.O[X.pr[X.no_of_pr - 1]].tot_deg();
	int n = 0;
	for(int i = X.A_d_maxdeg; i >= 0; i--) {
		if(X.A_d_coldeg[i] != 0) {
			if (n == 1) s << " + ";
			if(X.A_d_coldeg[i] == 1) {
				s << "P";
				if(l - i - 1 != 0)
					s << "(" << l - i - 1 << ")";
			}
			else {
				s << "P^" << X.A_d_coldeg[i];
				if(l - i - 1 != 0)
					s << "(" <<  l - i - 1 << ")";
			};
			n = 1;
		};
	};

	s << " --> ";

	n = 0;
	for(i = 0; i < X.sigma + 2; i++) {
		if(X.pr_deg[i] != 0) {
			if (n == 1) s << " + ";
			if(X.pr_deg[i] == 1) {
				s << "P";
				if(i != 1)
					s << "(" << i - 1 << ")";
			}
			else {
				s << "P^" << X.pr_deg[i];
				if(i != 1)
					s << "(" << i - 1 << ")";
			};
			n = 1;
		};
	};

	s << " --> w_R --> 0\n" << Line;
};

void disp_omegamat(ostream &s, scheme &X) {
	if(X.no_of_pr == 0)
		omega(X);
	int n, i;
	s << "the coloms of A_" << dim << " :\n";
	for(n = 0; n < X.A_d_cols; n++) {
		s << "[ ";
		counter = 2;
		for(i = 0; i < X.no_of_pr; i++) {
			s << X.A_d[i * X.A_d_cols + n];
			if(i != X.no_of_pr - 1)
				s << ";\n";
			counter = 0;
		};
		s << " ]\n";
	};
	s << Line;

	s << "w_R is generated by the following projections :\n";
	for(i = 0; i < X.no_of_pr; i++) {
		s << "pr_(" << X.O[X.pr[i]] << ")\n";
	};
	s << Line;
};

/*******************************************************************
linres
*******************************************************************/

void norm(polynomial *(*A), int A_c, int i, int jj) {
	K c = A[i][jj].coef[0];
	c.invert();
	for(int j = 0; j < A_c; j++)
		A[i][j] *= c;
};

void reduce_c(polynomial *(*A), int A_c, int k, int i, int jj) {
	// take care of Lc A[k][jj] = 1
	K c = A[i][jj].coef[0];
	for(int j = 0; j < A_c; j++)
		A[i][j] -= A[k][j] * c;
};

void interreduce(polynomial *(*A), int A_r, int A_c, int *Lt) {
	int i,j,k;

	for(i = 0; i < A_r; i++) {
		again:
		Lt[i] = -1;
		for(j = 0; j < A_c; j++) { // determine Lt
			if(sign(A[i][j]) == 0) continue;
			if(Lt[i] == -1 ) Lt[i] = j;
			else if(A[i][j].term[0] > A[i][Lt[i]].term[0])
				Lt[i] = j;
		};
		if(Lt[i] == -1) continue;
		for(k = 0; k < i; k++) // reduce
			if(Lt[k] == Lt[i])
				if(A[i][Lt[i]].term[0] == A[k][Lt[k]].term[0]) {
					reduce_c(A, A_c, k, i, Lt[i]);
					goto again;
				};
		norm(A, A_c, i, Lt[i]);
	}; // Lt's determined & interreduced & normed
};

void linid(scheme &X) {
	if(X.linid != 0) return;
	if (X.min_Ix == NULL)
		min_Ix(X);

	dim++;

	polynomial *(*A);
	polynomial *(*B);
	int A_r, A_c, B_r, B_c;
	int *Lt;
	int number;
	polynomial *tmp, t;
	monom m;
	K c;
	int Gb2;
	int *Gb2Lt;
	int *Gb2_B_r;
	polynomial *(*Gb2pol);

	int flag1 = 1;
	int i,j,k,l,L;
	int N = 0;

	for(i = 0; i < dim-1; i++)
		X.linres[0][i] = 0;

	A_c = 1;
	for(i = X.min_Ix->no_of_gens - 1; i >= 0; i--)
		if(X.min_Ix->gen[i].term[0].tot_deg() ==
			X.min_Ix->gen[0].term[0].tot_deg()) break;
	A_r = i + 1;// no of generators of lowest degree

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++) {
		A[i] = new polynomial[1];
		A[i][0] = X.min_Ix->gen[i];
	};
	Lt = new int[A_r];

	for(i = 0; i < A_r; i++) Lt[i] = 0;

	B_r = 0;
	for(i = 0; i < A_r - 1; i++)
		for(j = i + 1; j < A_r; j++) {
			m = lcm(A[j][Lt[j]].term[0], A[i][Lt[i]].term[0]);
			if(m.tot_deg() - A[i][0].term[0].tot_deg() == 1) B_r++;
		};

	X.linres[0][0] = A_r;
	if(verbose) cout << "I_X has " << A_r << " generators of degree "
							<< X.alpha << "\n";
	goto once;

entry:

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++)
		if(Lt[i] != -1) L++;

	X.linres[0][N] = L;
	if(verbose) cout << L << " syzygies in degree " << X.alpha - N << " found\n";

	// how many bi-syzygies
	B_r = 0;
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++)
			if(Lt[i] == Lt[j]) B_r++;
	};

	once:
	number = B_r;

	B_c = A_r;
	B = new polynomial*[B_r];
	for(i = 0; i < B_r; i++)
		B[i] = new polynomial[B_c];

	Gb2_B_r  = new int[B_r];
	Gb2Lt    = new int[B_r];
	Gb2pol   = new polynomial*[B_r];
	Gb2      = 0;

	tmp = new polynomial[A_c];

	// calc bi-syzygies
	B_r = 0;
	if(verbose && number != 0) cout << "eliminating syzygies ...\n";
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++) {
			if(Lt[i] != Lt[j]) continue;
			if(flag1 == 1) {
				m = lcm(A[j][0].term[0], A[i][0].term[0]);
				if(m.tot_deg() - A[i][0].term[0].tot_deg() > 1) continue;
				B[B_r][i] = m / A[i][0].term[0];
				B[B_r][j] = m / A[j][0].term[0];
			}
			else {
				B[B_r][i] = A[j][Lt[j]].term[0];
				B[B_r][j] = A[i][Lt[i]].term[0];
			};
			if(verbose) cout << "[" << number - B_r << "]";
			if(verbose) cout.flush();
			B[B_r][i].coef[0] = - B[B_r][i].coef[0];
			L = -1; // Lt(tmp)
			for(k = 0; k < A_c; k++)
				if(flag1 == 0) tmp[k] = A[j][k] * A[i][Lt[i]].term[0] -
												A[i][k] * A[j][Lt[j]].term[0];
				else tmp[k] = A[j][k] * (m / A[j][0].term[0]) -
								  A[i][k] * (m / A[i][0].term[0]);
			start:
			// determine Lt(tmp) & reduce tmp;
			L = -1;
			for(k = 0; k < A_c; k++) {
				if(sign(tmp[k]) == 0) continue;
				if(L == -1 ) L = k;
				else if(tmp[k].term[0] > tmp[L].term[0]) L = k;
			};
			if(L == -1) {
				B_r++;
				continue;
			};
			// no linear reduction possible
			for(k = 0; k < Gb2; k++) {
				if(L != Gb2Lt[k]) continue;
				if(tmp[L].term[0] != Gb2pol[k][Gb2Lt[k]].term[0]) continue;
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++)
					tmp[l] -= Gb2pol[k][l] * c;
				for(l = 0; l < B_c; l++)
					B[B_r][l] -= B[Gb2_B_r[k]][l] * c;
				goto start;
			};
			for(k = 0; k < A_r; k++) {
				if(Lt[k] != L) continue;
				if(! Zn_order_lt(A[k][Lt[k]].term[0],tmp[L].term[0])) continue;
				m = tmp[L].term[0] / A[k][Lt[k]].term[0];
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++) {
					t = A[k][l] * m;
					t *= c;
					tmp[l] -= t;
				};
				t = m;
				t *= c;
				B[B_r][k] -= t;
				goto start;
			};
			// new gbasis element of deg 2
			cout << ".";
			c = tmp[L].coef[0];
			c.invert();
			for(k = 0; k < B_c; k++)
				B[B_r][k] *= c;
			Gb2Lt[Gb2] = L;
			Gb2pol[Gb2] = new polynomial[A_c];
			for(k = 0; k < A_c; k++)
				Gb2pol[Gb2][k] = tmp[k] * c;
			Gb2_B_r[Gb2] = B_r;
			Gb2++;
			B_r++;
		};
	};
	if(verbose) cout << "\n";
	flag1 = 0;

	for(i = 0; i < Gb2; i++)
		for(j = 0; j < B_c; j++) B[Gb2_B_r[i]][j] = 0;
	delete[] Gb2_B_r;
	delete[] Gb2Lt;
	for(i = 0; i < Gb2; i++)
		delete[] Gb2pol[i];
	delete[] Gb2pol;

	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;
	// B -> A
	A = B;
	A_r = B_r;
	A_c = B_c;
	// clean up
	delete[] tmp;
	delete[] Lt;

	N++;

	if(A_r != 0) goto entry;
	// cleanup
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	X.linid = 1;
	dim--;
	if(verbose) cout << Line;
};

void linsyz(scheme &X) {
	if (X.min_Ix == NULL)
		min_Ix(X);

	dim++;

	polynomial *(*A);
	polynomial *(*B);
	int A_r, A_c, B_r, B_c;
	int *Lt;
	int number;
	polynomial *tmp, t;
	monom m;
	K c;
	int Gb2;
	int *Gb2Lt;
	int *Gb2_B_r;
	polynomial *(*Gb2pol);

	int flag1 = 1;
	int i,j,k,l,L;
	int N = 0;

	A_c = 1;
	for(i = X.min_Ix->no_of_gens - 1; i >= 0; i--)
		if(X.min_Ix->gen[i].term[0].tot_deg() ==
			X.min_Ix->gen[0].term[0].tot_deg()) break;
	A_r = i + 1;// no of generators of lowest degree

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++) {
		A[i] = new polynomial[1];
		A[i][0] = X.min_Ix->gen[i];
	};
	Lt = new int[A_r];

	for(i = 0; i < A_r; i++) Lt[i] = 0;

	B_r = 0;
	for(i = 0; i < A_r - 1; i++)
		for(j = i + 1; j < A_r; j++) {
			m = lcm(A[j][Lt[j]].term[0], A[i][Lt[i]].term[0]);
			if(m.tot_deg() - A[i][0].term[0].tot_deg() == 1) B_r++;
		};

	X.linres[0][0] = A_r;
	if(verbose) cout << "I_X has " << A_r << " generators of degree "
							<< X.alpha << "\n";
	goto once;

entry:

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++)
		if(Lt[i] != -1) L++;

	X.linres[0][N] = L;
	if(verbose) cout << L << " syzygies in degree " << X.alpha - N << " found\n";
	if(X.linres[0][N] != 0) {
		for(i = 0; i < A_r; i++) {
			if(Lt[i] == -1) continue;
			cout << "[ ";
			counter = 2;
			for(j = 0; j < A_c; j++) {
				cout << A[i][j];
				if(j != A_c - 1) cout << ";\n";
				counter = 0;
			};
			cout << " ]\n";
		};
	};
	goto end;

	once:
	number = B_r;

	B_c = A_r;
	B = new polynomial*[B_r];
	for(i = 0; i < B_r; i++)
		B[i] = new polynomial[B_c];

	Gb2_B_r  = new int[B_r];
	Gb2Lt    = new int[B_r];
	Gb2pol   = new polynomial*[B_r];
	Gb2      = 0;

	tmp = new polynomial[A_c];

	// calc bi-syzygies
	B_r = 0;
	if(verbose && number != 0) cout << "eliminating syzygies ...\n";
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++) {
			if(Lt[i] != Lt[j]) continue;
			if(flag1 == 1) {
				m = lcm(A[j][0].term[0], A[i][0].term[0]);
				if(m.tot_deg() - A[i][0].term[0].tot_deg() > 1) continue;
				B[B_r][i] = m / A[i][0].term[0];
				B[B_r][j] = m / A[j][0].term[0];
			}
			else {
				B[B_r][i] = A[j][Lt[j]].term[0];
				B[B_r][j] = A[i][Lt[i]].term[0];
			};
			if(verbose) cout << "[" << number - B_r << "]";
			if(verbose) cout.flush();
			B[B_r][i].coef[0] = - B[B_r][i].coef[0];
			L = -1; // Lt(tmp)
			for(k = 0; k < A_c; k++)
				if(flag1 == 0) tmp[k] = A[j][k] * A[i][Lt[i]].term[0] -
												A[i][k] * A[j][Lt[j]].term[0];
				else tmp[k] = A[j][k] * (m / A[j][0].term[0]) -
								  A[i][k] * (m / A[i][0].term[0]);
			start:
			// determine Lt(tmp) & reduce tmp;
			L = -1;
			for(k = 0; k < A_c; k++) {
				if(sign(tmp[k]) == 0) continue;
				if(L == -1 ) L = k;
				else if(tmp[k].term[0] > tmp[L].term[0]) L = k;
			};
			if(L == -1) {
				B_r++;
				continue;
			};
			for(k = 0; k < A_r; k++) {
				if(Lt[k] != L) continue;
				if(! Zn_order_lt(A[k][Lt[k]].term[0],tmp[L].term[0])) continue;
				m = tmp[L].term[0] / A[k][Lt[k]].term[0];
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++) {
					t = A[k][l] * m;
					t *= c;
					tmp[l] -= t;
				};
				t = m;
				t *= c;
				B[B_r][k] -= t;
				goto start;
			};
			// no linear reduction possible
			for(k = 0; k < Gb2; k++) {
				if(L != Gb2Lt[k]) continue;
				if(tmp[L].term[0] != Gb2pol[k][Gb2Lt[k]].term[0]) continue;
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++)
					tmp[l] -= Gb2pol[k][l] * c;
				for(l = 0; l < B_c; l++)
					B[B_r][l] -= B[Gb2_B_r[k]][l] * c;
				goto start;
			};
			// new gbasis element of deg 2
			cout << ".";
			c = tmp[L].coef[0];
			c.invert();
			for(k = 0; k < B_c; k++)
				B[B_r][k] *= c;
			Gb2Lt[Gb2] = L;
			Gb2pol[Gb2] = new polynomial[A_c];
			for(k = 0; k < A_c; k++)
				Gb2pol[Gb2][k] = tmp[k] * c;
			Gb2_B_r[Gb2] = B_r;
			Gb2++;
			B_r++;
		};
	};
	if(verbose) cout << "\n";
	flag1 = 0;

	for(i = 0; i < Gb2; i++)
		for(j = 0; j < B_c; j++) B[Gb2_B_r[i]][j] = 0;
	delete[] Gb2_B_r;
	delete[] Gb2Lt;
	for(i = 0; i < Gb2; i++)
		delete[] Gb2pol[i];
	delete[] Gb2pol;

	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;
	// B -> A
	A = B;
	A_r = B_r;
	A_c = B_c;
	// clean up
	delete[] tmp;
	delete[] Lt;

	N++;

	if(A_r != 0) goto entry;
	// cleanup
end:
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	dim--;

	if(verbose) cout << Line;
};

void putlinsyz(scheme &X) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	if (X.min_Ix == NULL)
		min_Ix(X);

	dim++;

	polynomial *(*A);
	polynomial *(*B);
	int A_r, A_c, B_r, B_c;
	int *Lt;
	int number;
	polynomial *tmp, t;
	monom m;
	K c;
	int Gb2;
	int *Gb2Lt;
	int *Gb2_B_r;
	polynomial *(*Gb2pol);

	int flag1 = 1;
	int i,j,k,l,L;
	int N = 0;

	A_c = 1;
	for(i = X.min_Ix->no_of_gens - 1; i >= 0; i--)
		if(X.min_Ix->gen[i].term[0].tot_deg() ==
			X.min_Ix->gen[0].term[0].tot_deg()) break;
	A_r = i + 1;// no of generators of lowest degree

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++) {
		A[i] = new polynomial[1];
		A[i][0] = X.min_Ix->gen[i];
	};
	Lt = new int[A_r];

	for(i = 0; i < A_r; i++) Lt[i] = 0;

	B_r = 0;
	for(i = 0; i < A_r - 1; i++)
		for(j = i + 1; j < A_r; j++) {
			m = lcm(A[j][Lt[j]].term[0], A[i][Lt[i]].term[0]);
			if(m.tot_deg() - A[i][0].term[0].tot_deg() == 1) B_r++;
		};

	X.linres[0][0] = A_r;
	if(verbose) cout << "I_X has " << A_r << " generators of degree "
							<< X.alpha << "\n";
	goto once;

entry:

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++)
		if(Lt[i] != -1) L++;

	X.linres[0][N] = L;
	if(verbose) cout << L << " syzygies in degree " << X.alpha - N << " found\n";
	if(X.linres[0][N] != 0) {
		to << A_c << "\n" << X.linres[0][N] << "\n";
		for(i = 0; i < A_r; i++) {
			if(Lt[i] == -1) continue;
			for(j = 0; j < A_c; j++) {
				A[i][j].put(to);
				to << "\n";
			};
		};
	};
	goto end;

	once:
	number = B_r;

	B_c = A_r;
	B = new polynomial*[B_r];
	for(i = 0; i < B_r; i++)
		B[i] = new polynomial[B_c];

	Gb2_B_r  = new int[B_r];
	Gb2Lt    = new int[B_r];
	Gb2pol   = new polynomial*[B_r];
	Gb2      = 0;

	tmp = new polynomial[A_c];

	// calc bi-syzygies
	B_r = 0;
	if(verbose && number != 0) cout << "eliminating syzygies ...\n";
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++) {
			if(Lt[i] != Lt[j]) continue;
			if(flag1 == 1) {
				m = lcm(A[j][0].term[0], A[i][0].term[0]);
				if(m.tot_deg() - A[i][0].term[0].tot_deg() > 1) continue;
				B[B_r][i] = m / A[i][0].term[0];
				B[B_r][j] = m / A[j][0].term[0];
			}
			else {
				B[B_r][i] = A[j][Lt[j]].term[0];
				B[B_r][j] = A[i][Lt[i]].term[0];
			};
			if(verbose) cout << "[" << number - B_r << "]";
			if(verbose) cout.flush();
			B[B_r][i].coef[0] = - B[B_r][i].coef[0];
			L = -1; // Lt(tmp)
			for(k = 0; k < A_c; k++)
				if(flag1 == 0) tmp[k] = A[j][k] * A[i][Lt[i]].term[0] -
												A[i][k] * A[j][Lt[j]].term[0];
				else tmp[k] = A[j][k] * (m / A[j][0].term[0]) -
								  A[i][k] * (m / A[i][0].term[0]);
			start:
			// determine Lt(tmp) & reduce tmp;
			L = -1;
			for(k = 0; k < A_c; k++) {
				if(sign(tmp[k]) == 0) continue;
				if(L == -1 ) L = k;
				else if(tmp[k].term[0] > tmp[L].term[0]) L = k;
			};
			if(L == -1) {
				B_r++;
				continue;
			};
			for(k = 0; k < A_r; k++) {
				if(Lt[k] != L) continue;
				if(! Zn_order_lt(A[k][Lt[k]].term[0],tmp[L].term[0])) continue;
				m = tmp[L].term[0] / A[k][Lt[k]].term[0];
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++) {
					t = A[k][l] * m;
					t *= c;
					tmp[l] -= t;
				};
				t = m;
				t *= c;
				B[B_r][k] -= t;
				goto start;
			};
			// no linear reduction possible
			for(k = 0; k < Gb2; k++) {
				if(L != Gb2Lt[k]) continue;
				if(tmp[L].term[0] != Gb2pol[k][Gb2Lt[k]].term[0]) continue;
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++)
					tmp[l] -= Gb2pol[k][l] * c;
				for(l = 0; l < B_c; l++)
					B[B_r][l] -= B[Gb2_B_r[k]][l] * c;
				goto start;
			};
			// new gbasis element of deg 2
			cout << ".";
			c = tmp[L].coef[0];
			c.invert();
			for(k = 0; k < B_c; k++)
				B[B_r][k] *= c;
			Gb2Lt[Gb2] = L;
			Gb2pol[Gb2] = new polynomial[A_c];
			for(k = 0; k < A_c; k++)
				Gb2pol[Gb2][k] = tmp[k] * c;
			Gb2_B_r[Gb2] = B_r;
			Gb2++;
			B_r++;
		};
	};
	if(verbose) cout << "\n";
	flag1 = 0;

	for(i = 0; i < Gb2; i++)
		for(j = 0; j < B_c; j++) B[Gb2_B_r[i]][j] = 0;
	delete[] Gb2_B_r;
	delete[] Gb2Lt;
	for(i = 0; i < Gb2; i++)
		delete[] Gb2pol[i];
	delete[] Gb2pol;

	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;
	// B -> A
	A = B;
	A_r = B_r;
	A_c = B_c;
	// clean up
	delete[] tmp;
	delete[] Lt;

	N++;

	if(A_r != 0) goto entry;
	// cleanup
end:
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	dim--;

	if(verbose) cout << Line;
};


void putlinid(scheme &X, char *file_name, int mode) {
	ofstream to(file_name, mode);
	if(!to)  {
		error(30);
		return;
	};

	if (X.min_Ix == NULL)
		min_Ix(X);

	dim++;

	polynomial *(*A);
	polynomial *(*B);
	int A_r, A_c, B_r, B_c;
	int *Lt;
	int number;
	polynomial *tmp, t;
	monom m;
	K c;
	int Gb2;
	int *Gb2Lt;
	int *Gb2_B_r;
	polynomial *(*Gb2pol);

	int flag1 = 1;
	int i,j,k,l,L;
	int N = 0;

	for(i = 0; i < dim-1; i++)
		X.linres[0][i] = 0;

	A_c = 1;
	for(i = X.min_Ix->no_of_gens - 1; i >= 0; i--)
		if(X.min_Ix->gen[i].term[0].tot_deg() ==
			X.min_Ix->gen[0].term[0].tot_deg()) break;
	A_r = i + 1;// no of generators of lowest degree

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++) {
		A[i] = new polynomial[1];
		A[i][0] = X.min_Ix->gen[i];
	};
	Lt = new int[A_r];

	for(i = 0; i < A_r; i++) Lt[i] = 0;

	B_r = 0;
	for(i = 0; i < A_r - 1; i++)
		for(j = i + 1; j < A_r; j++) {
			m = lcm(A[j][Lt[j]].term[0], A[i][Lt[i]].term[0]);
			if(m.tot_deg() - A[i][0].term[0].tot_deg() == 1) B_r++;
		};

	X.linres[0][0] = A_r;
	if(verbose) cout << "I_X has " << A_r << " generators of degree "
							<< X.alpha << "\n";
	goto once;

entry:

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++)
		if(Lt[i] != -1) L++;

	X.linres[0][N] = L;
	if(verbose) cout << L << " syzygies in degree " << X.alpha - N << " found\n";

	// how many bi-syzygies
	B_r = 0;
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++)
			if(Lt[i] == Lt[j]) B_r++;
	};

	once:
	number = B_r;

	B_c = A_r;
	B = new polynomial*[B_r];
	for(i = 0; i < B_r; i++)
		B[i] = new polynomial[B_c];

	Gb2_B_r  = new int[B_r];
	Gb2Lt    = new int[B_r];
	Gb2pol   = new polynomial*[B_r];
	Gb2      = 0;

	tmp = new polynomial[A_c];

	if(X.linres[0][N] != 0) {
		to << "\n" << N + 1 << ". matrix of linear resolution of I_X : \n";
		to << A_c << "\n" << X.linres[0][N] << "\n";
		for(i = 0; i < A_r; i++) {
			if(Lt[i] == -1) continue;
			for(j = 0; j < A_c; j++) {
				A[i][j].put(to);
				to << "\n";
			};
		};
	};
	// calc bi-syzygies
	B_r = 0;
	if(verbose && number != 0) cout << "eliminating syzygies ...\n";
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++) {
			if(Lt[i] != Lt[j]) continue;
			if(flag1 == 1) {
				m = lcm(A[j][0].term[0], A[i][0].term[0]);
				if(m.tot_deg() - A[i][0].term[0].tot_deg() > 1) continue;
				B[B_r][i] = m / A[i][0].term[0];
				B[B_r][j] = m / A[j][0].term[0];
			}
			else {
				B[B_r][i] = A[j][Lt[j]].term[0];
				B[B_r][j] = A[i][Lt[i]].term[0];
			};
			if(verbose) cout << "[" << number - B_r << "]";
			if(verbose) cout.flush();
			B[B_r][i].coef[0] = - B[B_r][i].coef[0];
			L = -1; // Lt(tmp)
			for(k = 0; k < A_c; k++)
				if(flag1 == 0) tmp[k] = A[j][k] * A[i][Lt[i]].term[0] -
												A[i][k] * A[j][Lt[j]].term[0];
				else tmp[k] = A[j][k] * (m / A[j][0].term[0]) -
								  A[i][k] * (m / A[i][0].term[0]);
			start:
			// determine Lt(tmp) & reduce tmp;
			L = -1;
			for(k = 0; k < A_c; k++) {
				if(sign(tmp[k]) == 0) continue;
				if(L == -1 ) L = k;
				else if(tmp[k].term[0] > tmp[L].term[0]) L = k;
			};
			if(L == -1) {
				B_r++;
				continue;
			};
			for(k = 0; k < A_r; k++) {
				if(Lt[k] != L) continue;
				if(! Zn_order_lt(A[k][Lt[k]].term[0],tmp[L].term[0])) continue;
				m = tmp[L].term[0] / A[k][Lt[k]].term[0];
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++) {
					t = A[k][l] * m;
					t *= c;
					tmp[l] -= t;
				};
				t = m;
				t *= c;
				B[B_r][k] -= t;
				goto start;
			};
			// no linear reduction possible
			for(k = 0; k < Gb2; k++) {
				if(L != Gb2Lt[k]) continue;
				if(tmp[L].term[0] != Gb2pol[k][Gb2Lt[k]].term[0]) continue;
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++)
					tmp[l] -= Gb2pol[k][l] * c;
				for(l = 0; l < B_c; l++)
					B[B_r][l] -= B[Gb2_B_r[k]][l] * c;
				goto start;
			};
			// new gbasis element of deg 2
			cout << ".";
			c = tmp[L].coef[0];
			c.invert();
			for(k = 0; k < B_c; k++)
				B[B_r][k] *= c;
			Gb2Lt[Gb2] = L;
			Gb2pol[Gb2] = new polynomial[A_c];
			for(k = 0; k < A_c; k++)
				Gb2pol[Gb2][k] = tmp[k] * c;
			Gb2_B_r[Gb2] = B_r;
			Gb2++;
			B_r++;
		};
	};
	if(verbose) cout << "\n";
	flag1 = 0;

	for(i = 0; i < Gb2; i++)
		for(j = 0; j < B_c; j++) B[Gb2_B_r[i]][j] = 0;
	delete[] Gb2_B_r;
	delete[] Gb2Lt;
	for(i = 0; i < Gb2; i++)
		delete[] Gb2pol[i];
	delete[] Gb2pol;

	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;
	// B -> A
	A = B;
	A_r = B_r;
	A_c = B_c;
	// clean up
	delete[] tmp;
	delete[] Lt;

	N++;

	if(A_r != 0) goto entry;
	// cleanup
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	X.linid = 1;
	dim--;

	if(verbose) cout << Line;
};

/*******************************************************************
linom
*******************************************************************/

void linom(scheme &X) {
	if(X.linom != 0) return;
	if(X.no_of_pr == 0)
		omega(X);

	dim++;

	polynomial *(*A);
	polynomial *(*B);
	int A_r, A_c, B_r, B_c;
	int *Lt;
	int number;
	polynomial *tmp, t;
	monom m;
	K c;
	int Gb2;
	int *Gb2Lt;
	int *Gb2_B_r;
	polynomial *(*Gb2pol);

	int i,j,k,l,L;
	int N = 0;

	for(i = 0; i < dim - 1; i++)
		X.linres[1][i] = 0;

	A_c = 1; // no of projektions of lowest degree
	for(i = 0; i < X.no_of_pr - 1; i++)
		if(X.O[X.pr[i]].tot_deg() == X.O[X.pr[X.no_of_pr-1]].tot_deg()) A_c++;

	A_r = 0; // no of lin syz
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				A_r++;
				break;
			};
		};
	};

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++)
		A[i] = new polynomial[A_c];

	A_r = 0;
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				for(k = 0; k < A_c; k++)
					A[A_r][k] = X.A_d[(X.no_of_pr - A_c + k) * X.A_d_cols + i];
				A_r++;
				break;
			};
		};
	}; // A created
	X.linres[1][dim - 2] = A_c;
	if(verbose) cout << "w_R has " << A_c << " generators in degree "
						  << -X.sigma << "\n";
	N++;

entry:

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++) if(Lt[i] != -1) L++;

	X.linres[1][dim - N - 2] = L;
	if(verbose) cout << L << " syzygies in degree " << - X.sigma - N << " found\n";

	// how many bi-syzygies
	B_r = 0;
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++)
			if(Lt[i] == Lt[j]) B_r++;
	};
	number = B_r;

	B_c = A_r;
	B = new polynomial*[B_r];
	for(i = 0; i < B_r; i++)
		B[i] = new polynomial[B_c];

	Gb2_B_r  = new int[B_r];
	Gb2Lt    = new int[B_r];
	Gb2pol   = new polynomial*[B_r];
	Gb2      = 0;

	tmp = new polynomial[A_c];
	// calc bi-syzygies
	B_r = 0;
	if(verbose && number != 0) cout << "eliminating syzygies ...\n";
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++) {
			if(Lt[i] != Lt[j]) continue;
			if(verbose) cout << "[" << number - B_r << "]";
			if(verbose) cout.flush();
			B[B_r][i] = A[j][Lt[j]].term[0];
			B[B_r][j] = A[i][Lt[i]].term[0];
			B[B_r][i].coef[0] = - B[B_r][i].coef[0];
			L = -1; // Lt(tmp)
			for(k = 0; k < A_c; k++)
				tmp[k] = A[j][k] * A[i][Lt[i]].term[0] -
							A[i][k] * A[j][Lt[j]].term[0];
			start:
			// determine Lt(tmp) & reduce tmp;
			L = -1;
			for(k = 0; k < A_c; k++) {
				if(sign(tmp[k]) == 0) continue;
				if(L == -1 ) L = k;
				else if(tmp[k].term[0] > tmp[L].term[0]) L = k;
			};
			if(L == -1) {
				B_r++;
				continue;
			};
			// no linear reduction possible
			for(k = 0; k < Gb2; k++) {
				if(L != Gb2Lt[k]) continue;
				if(tmp[L].term[0] != Gb2pol[k][Gb2Lt[k]].term[0]) continue;
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++)
					tmp[l] -= Gb2pol[k][l] * c;
				for(l = 0; l < B_c; l++)
					B[B_r][l] -= B[Gb2_B_r[k]][l] * c;
				goto start;
			};
			for(k = 0; k < A_r; k++) {
				if(Lt[k] != L) continue;
				if(! Zn_order_lt(A[k][Lt[k]].term[0],tmp[L].term[0])) continue;
				m = tmp[L].term[0] / A[k][Lt[k]].term[0];
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++) {
					t = A[k][l] * m;
					t *= c;
					tmp[l] -= t;
				};
				t = m;
				t *= c;
				B[B_r][k] -= t;
				goto start;
			};
			// new gbasis element of deg 2
			cout << ".";
			c = tmp[L].coef[0];
			c.invert();
			for(k = 0; k < B_c; k++)
				B[B_r][k] *= c;
			Gb2Lt[Gb2] = L;
			Gb2pol[Gb2] = new polynomial[A_c];
			for(k = 0; k < A_c; k++)
				Gb2pol[Gb2][k] = tmp[k] * c;
			Gb2_B_r[Gb2] = B_r;
			Gb2++;
			B_r++;
		};
	};
	if(verbose) cout << "\n";

	for(i = 0; i < Gb2; i++)
		for(j = 0; j < B_c; j++) B[Gb2_B_r[i]][j] = 0;
	delete[] Gb2_B_r;
	delete[] Gb2Lt;
	for(i = 0; i < Gb2; i++)
		delete[] Gb2pol[i];
	delete[] Gb2pol;

	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;
	// B -> A
	A = B;
	A_r = B_r;
	A_c = B_c;
	// clean up
	delete[] tmp;
	delete[] Lt;

	N++;

	if(A_r != 0) goto entry;

	// cleanup
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	X.linom = 1;
	dim--;
	if(verbose) cout << Line;
};

void linomsyz(scheme &X) {
	if(X.no_of_pr == 0)
		omega(X);

	dim++;

	polynomial *(*A);
	int A_r, A_c;
	int *Lt;
	int i,j,k,L;
	int N = 0;

	A_c = 1; // no of projektions of lowest degree
	for(i = 0; i < X.no_of_pr - 1; i++)
		if(X.O[X.pr[i]].tot_deg() == X.O[X.pr[X.no_of_pr-1]].tot_deg()) A_c++;

	A_r = 0; // no of lin syz
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				A_r++;
				break;
			};
		};
	};

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++)
		A[i] = new polynomial[A_c];

	A_r = 0;
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				for(k = 0; k < A_c; k++)
					A[A_r][k] = X.A_d[(X.no_of_pr - A_c + k) * X.A_d_cols + i];
				A_r++;
				break;
			};
		};
	}; // A created
	X.linres[1][dim - 2] = A_c;
	if(verbose) cout << "w_R has " << A_c << " generators in degree "
						  << -X.sigma << "\n";
	N++;

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++)
		if(Lt[i] != -1) L++;

	X.linres[1][dim - N - 2] = L;
	if(verbose) cout << L << " syzygies in degree " << - X.sigma - N << " found\n";


	if(X.linres[1][dim - N - 2] != 0) {
		for(i = 0; i < A_r; i++) {
			if(Lt[i] == -1) continue;
			cout << "[ ";
			counter = 2;
			for(j = 0; j < A_c; j++) {
				cout << A[i][j];
				if(j != A_c - 1) cout << ";\n";
				counter = 0;
			};
			cout << " ]\n";
		};

	};


	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	delete[] Lt;

	dim--;
	if(verbose) cout << Line;
};

void putlinomsyz(scheme &X) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};

	if(X.no_of_pr == 0)
		omega(X);

	dim++;

	polynomial *(*A);
	int A_r, A_c;
	int *Lt;

	int i,j,k,L;
	int N = 0;

	A_c = 1; // no of projektions of lowest degree
	for(i = 0; i < X.no_of_pr - 1; i++)
		if(X.O[X.pr[i]].tot_deg() == X.O[X.pr[X.no_of_pr-1]].tot_deg()) A_c++;

	A_r = 0; // no of lin syz
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				A_r++;
				break;
			};
		};
	};

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++)
		A[i] = new polynomial[A_c];

	A_r = 0;
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				for(k = 0; k < A_c; k++)
					A[A_r][k] = X.A_d[(X.no_of_pr - A_c + k) * X.A_d_cols + i];
				A_r++;
				break;
			};
		};
	}; // A created
	X.linres[1][dim - 2] = A_c;
	if(verbose) cout << "w_R has " << A_c << " generators in degree "
						  << -X.sigma << "\n";
	N++;

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++)
		if(Lt[i] != -1) L++;

	X.linres[1][dim - N - 2] = L;
	if(verbose) cout << L << " syzygies in degree " << - X.sigma - N << " found\n";


	if(X.linres[1][dim - N - 2] != 0) {
		to << A_c << "\n" << X.linres[1][dim - N - 2] << "\n";
		for(i = 0; i < A_r; i++) {
			if(Lt[i] == -1) continue;
			for(j = 0; j < A_c; j++) {
				A[i][j].put(to);
				to << "\n";
			};
		};
	};


	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	delete[] Lt;

	dim--;
	if(verbose) cout << Line;
};

void putlinom(scheme &X, char *file_name, int mode) {
	ofstream to(file_name, mode);
	if(!to)  {
		error(30);
		return;
	};

	if(X.no_of_pr == 0)
		omega(X);

	dim++;

	polynomial *(*A);
	polynomial *(*B);
	int A_r, A_c, B_r, B_c;
	int *Lt;
	int number;
	polynomial *tmp, t;
	monom m;
	K c;
	int Gb2;
	int *Gb2Lt;
	int *Gb2_B_r;
	polynomial *(*Gb2pol);

	int i,j,k,l,L;
	int N = 0;

	for(i = 0; i < dim - 1; i++)
		X.linres[1][i] = 0;

	A_c = 1; // no of projektions of lowest degree
	for(i = 0; i < X.no_of_pr - 1; i++)
		if(X.O[X.pr[i]].tot_deg() == X.O[X.pr[X.no_of_pr-1]].tot_deg()) A_c++;

	A_r = 0; // no of lin syz
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				A_r++;
				break;
			};
		};
	};

	A = new polynomial*[A_r];
	for(i = 0; i < A_r; i++)
		A[i] = new polynomial[A_c];

	A_r = 0;
	for(i = 0; i < X.A_d_cols; i++) {
		for(j = X.no_of_pr - A_c; j < X.no_of_pr; j++) {
			if(X.A_d[j * X.A_d_cols + i].term[0].tot_deg() +
				X.A_d[j * X.A_d_cols + i].term[0].exp[proj-1] == 1) {
				for(k = 0; k < A_c; k++)
					A[A_r][k] = X.A_d[(X.no_of_pr - A_c + k) * X.A_d_cols + i];
				A_r++;
				break;
			};
		};
	}; // A created
	X.linres[1][dim - 2] = A_c;
	if(verbose) cout << "w_R has " << A_c << " generators in degree "
						  << -X.sigma << "\n";
	N++;

entry:

	Lt = new int[A_r];

	interreduce(A, A_r, A_c, Lt);

	L = 0;
	for(i = 0; i < A_r; i++)
		if(Lt[i] != -1) L++;

	X.linres[1][dim - N - 2] = L;
	if(verbose) cout << L << " syzygies in degree " << - X.sigma - N << " found\n";

	// how many bi-syzygies
	B_r = 0;
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++)
			if(Lt[i] == Lt[j]) B_r++;
	};
	number = B_r;

	B_c = A_r;
	B = new polynomial*[B_r];
	for(i = 0; i < B_r; i++)
		B[i] = new polynomial[B_c];

	Gb2_B_r  = new int[B_r];
	Gb2Lt    = new int[B_r];
	Gb2pol   = new polynomial*[B_r];
	Gb2      = 0;

	tmp = new polynomial[A_c];

	if(X.linres[1][dim - N - 2] != 0) {
		to << "\n" << N << ". matrix of linear resolution of w_R : \n";
		to << A_c << "\n" << X.linres[1][dim - N - 2] << "\n";
		for(i = 0; i < A_r; i++) {
			if(Lt[i] == -1) continue;
			for(j = 0; j < A_c; j++) {
				A[i][j].put(to);
				to << "\n";
			};
		};
	};
	// calc bi-syzygies
	B_r = 0;
	if(verbose) cout << "eliminating syzygies ...\n";
	for(i = 0; i < A_r - 1; i++) {
		if(Lt[i] == -1) continue;
		for(j = i + 1; j < A_r; j++) {
			if(Lt[i] != Lt[j]) continue;
			if(verbose) cout << "[" << number - B_r << "]";
			if(verbose) cout.flush();
			B[B_r][i] = A[j][Lt[j]].term[0];
			B[B_r][j] = A[i][Lt[i]].term[0];
			B[B_r][i].coef[0] = - B[B_r][i].coef[0];
			L = -1; // Lt(tmp)
			for(k = 0; k < A_c; k++)
				tmp[k] = A[j][k] * A[i][Lt[i]].term[0] -
							A[i][k] * A[j][Lt[j]].term[0];
			start:
			// determine Lt(tmp) & reduce tmp;
			L = -1;
			for(k = 0; k < A_c; k++) {
				if(sign(tmp[k]) == 0) continue;
				if(L == -1 ) L = k;
				else if(tmp[k].term[0] > tmp[L].term[0]) L = k;
			};
			if(L == -1) {
				B_r++;
				continue;
			};
			for(k = 0; k < A_r; k++) {
				if(Lt[k] != L) continue;
				if(! Zn_order_lt(A[k][Lt[k]].term[0],tmp[L].term[0])) continue;
				m = tmp[L].term[0] / A[k][Lt[k]].term[0];
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++) {
					t = A[k][l] * m;
					t *= c;
					tmp[l] -= t;
				};
				t = m;
				t *= c;
				B[B_r][k] -= t;
				goto start;
			};
			// no linerar reduction possible
			for(k = 0; k < Gb2; k++) {
				if(L != Gb2Lt[k]) continue;
				if(tmp[L].term[0] != Gb2pol[k][Gb2Lt[k]].term[0]) continue;
				c = tmp[L].coef[0];
				for(l = 0; l < A_c; l++)
					tmp[l] -= Gb2pol[k][l] * c;
				for(l = 0; l < B_c; l++)
					B[B_r][l] -= B[Gb2_B_r[k]][l] * c;
				goto start;
			};
			// new gbasis element of deg 2
			cout << ".";
			c = tmp[L].coef[0];
			c.invert();
			for(k = 0; k < B_c; k++)
				B[B_r][k] *= c;
			Gb2Lt[Gb2] = L;
			Gb2pol[Gb2] = new polynomial[A_c];
			for(k = 0; k < A_c; k++)
				Gb2pol[Gb2][k] = tmp[k] * c;
			Gb2_B_r[Gb2] = B_r;
			Gb2++;
			B_r++;
		};
	};
	if(verbose) cout << "\n";

	for(i = 0; i < Gb2; i++)
		for(j = 0; j < B_c; j++) B[Gb2_B_r[i]][j] = 0;
	delete[] Gb2_B_r;
	delete[] Gb2Lt;
	for(i = 0; i < Gb2; i++)
		delete[] Gb2pol[i];
	delete[] Gb2pol;

	// delete A
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;
	// B -> A
	A = B;
	A_r = B_r;
	A_c = B_c;
	// clean up
	delete[] tmp;
	delete[] Lt;

	N++;

	if(A_r != 0) goto entry;

	// cleanup
	for(i = 0; i < A_r; i++)
		delete[] A[i];
	delete[] A;

	X.linom = 1;
	dim--;
	if(verbose) cout << Line;
};

void disp_linid(ostream &s, scheme &X) {
	int i,j,k;
	int *l = new int[dim];
	s << "    | ";
	k = 6;
	for(i = 0; i < dim ; i++) {
		l[i] = max(3, len(X.linres[0][i]));
		s << i+1;
		for(j = -1; j < l[i] - len(i+1); j++) s << " ";
		k += l[i] + 1;
	};
	s << "\n";
	for(j = 0; j < k; j++) s << "-";
	s << "\n";
	s << "deg | ";
	for(i = 0; i < dim ; i++) {
		s << X.alpha - i;
		for(j = -1; j < l[i] - len(X.alpha - i); j++) s << " ";
	};
	s << "\n";
	s << "I_X | ";
	for(i = 0; i < dim ; i++) {
		s << X.linres[0][i];
		for(j = -1; j < l[i] - len(X.linres[0][i]); j++) s << " ";
	};
	s << "\n";
	s << Line;
	delete[] l;
};

void disp_linom(ostream &s, scheme &X) {
	int i,j,k;
	int *l = new int[dim];
	s << "    | ";
	k = 6;
	for(i = 0; i < dim ; i++) {
		l[i] = max(3, len(X.linres[1][dim - i -1]));
		s << i+1;
		for(j = -1; j < l[i] - len(i+1); j++) s << " ";
		k += l[i] + 1;
	};
	s << "\n";
	for(j = 0; j < k; j++) s << "-";
	s << "\n";
	s << "deg | ";
	for(i = 0; i < dim; i++) {
		s << - X.sigma - i;
		for(j = -1; j < l[i] - len(- X.sigma - i); j++) s << " ";
	};
	s << "\n";
	s << "w_R | ";
	for(i = dim - 1; i >= 0; i--) {
		s << X.linres[1][i];
		for(j = -1; j < l[i] - len(X.linres[1][i]); j++) s << " ";
	};
	s << "\n";
	s << Line;
	delete[] l;
};

void disp_linres(ostream &s, scheme &X) {
	int i,j,k;
	int *l = new int[dim];
	s << "    | ";
	k = 6;
	for(i = 0; i < dim ; i++) {
		l[i] = max(3, len(X.linres[0][i]));
		l[i] = max(l[i], len(X.linres[1][i]));
		s << i+1;
		for(j = -1; j < l[i] - len(i+1); j++) s << " ";
		k += l[i] + 1;
	};
	s << "\n";
	for(j = 0; j < k; j++) s << "-";
	s << "\n";
	s << "deg | ";
	for(i = 0; i < dim ; i++) {
		s << X.alpha - i;
		for(j = -1; j < l[i] - len(X.alpha - i); j++) s << " ";
	};
	s << "\n";
	s << "I_X | ";
	for(i = 0; i < dim ; i++) {
		s << X.linres[0][i];
		for(j = -1; j < l[i] - len(X.linres[0][i]); j++) s << " ";
	};
	s << "\n";
	for(j = 0; j < k; j++) s << "-";
	s << "\n";
	s << "w_R | ";
	for(i = 0; i < dim ; i++) {
		s << X.linres[1][i];
		for(j = -1; j < l[i] - len(X.linres[1][i]); j++) s << " ";
	};
	s << "\n";
	s << "deg | ";
	for(i = 0; i < dim ; i++) {
		s << - X.sigma - (dim - i) + 1;
		for(j = -1; j < l[i] - len(- X.sigma - (dim - i) + 1); j++) s << " ";
	};
	s << "\n";
	s << Line;
	delete[] l;
};

void writelinid(scheme &X) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	disp_linid(to, X);
};

void writelinom(scheme &X) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	disp_linom(to, X);
};

void writelinres(scheme &X) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	disp_linres(to, X);
};

/*******************************************************************
create
*******************************************************************/

void read_var(int &v, char *cc);

int read_scheme(ifstream &in, int v) {
	int rows, cols;
	in >> rows >> cols;
	if (rows != proj) {
		error(31);
		return 1;
	};
	var[v].var_ptr = new scheme(cols);
	matrix *M = ((scheme *)var[v].var_ptr)->P;
	K tmp;
	int i, j;
	for(i = 0; i < M->rows; i++) {
		in >> tmp;
		if(tmp == 0) {
			cerr << "first coordinate is expected to be different from zero\n";
			cerr << "first coordinate as '1'\n";
			tmp = 1;
		};
		for(j = 0; j < M->cols; j++) {
			in >> M->a[i][j];
			M->a[i][j] /= tmp;
		};
	};
	return 0;
};

void get_scheme(void) {
	int i, j;
	char cc[13];
	int v;
	cin.width(13);
	cin >> cc;

	for (i = 0; i < strlen(cc); i++)
		if (! isalnum((int) cc[i])) {
			error(105);
			return;
		};

	read_var(v, cc);
	if(v != -2) {
		error(111);  // par name already exists
		return;
	};

	for (i = 0; no_of_vars; i++) {
		if (var[i].name[0] == '\0') {
			strcpy(var[i].name, cc);
			var[i].typ  = 3;
			v = i;
			break;
		};
	};

	if (i == no_of_vars) {
		error(110);
		return;
	};

	cout << "enter number of points : ";
	int rows, cols;
	rows = dim;
	cin >> cols;
	var[v].var_ptr = new scheme(cols);
	matrix *M = ((scheme *)var[v].var_ptr)->P;
	K tmp;
	char c;
	for(i = 0; i < M->rows; i++) {
		tmp = 0;
		cin >> tmp;
		if(tmp == 0) {
			cerr << "first coordinate is expected to be different from zero\n";
			cerr << "first coordinate treated as '1'\n";
			tmp = 1;
		};
		cin.get(c);
		for(j = 0; j < M->cols; j++) {
			M->a[i][j] = 0;
			cin >> M->a[i][j];
			M->a[i][j] /= tmp;
		};
		cout << i + 1 << ". point entered : " << "<1";
		for(j = 0; j < M->cols; j++)
			cout << ":" << M->a[i][j];
		cout << ">\n";
	};
};

void create(int typ) {
	int i;
	int v;
	char cc[13];

	cin.width(13);
	cin >> cc;

	for (i = 0; i < strlen(cc); i++)
		if (! isalnum((int) cc[i])) {
			error(105);
			return;
		};

	read_var(v, cc);
	if(v != -2) {
		error(111);  // par name already exists
		return;
	};

	for (i = 0; no_of_vars; i++) {
		if (var[i].name[0] == '\0') {
			strcpy(var[i].name, cc);
			var[i].typ  = typ;
			v = i;
			break;
		};
	};

	if (i == no_of_vars) {
		error(110);
		return;
	};

	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ifstream from(file_name);
	if(!from)  {
		error(30);
		var[i].name[0] = '\0';
		return;
	};
	switch (typ) {
		case 3 : // scheme
			if(read_scheme(from, v))
				var[v].name[0] = '\0';
			break;
	};
	cout << Line;
};

void putscheme(int v) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	to << proj << "\n";
	to << ((scheme *)var[v].var_ptr)->s << "\n";
	matrix *M = ((scheme *)var[v].var_ptr)->P;
	int i, j;
	for(i = 0; i < M->rows; i++) {
		to << 1 << "\n";
		for(j = 0; j < M->cols; j++)
			to << M->a[i][j] << "\n";
	};
	cout << Line;
};

void putbeta(int v) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	if(((scheme *)var[v].var_ptr)->beta_matrix == 0)
		beta_matrix(*((scheme *)var[v].var_ptr));
	else
		cout << Line;

	matrix *M = ((scheme *)var[v].var_ptr)->beta_matrix;
	to << M->cols << "\n";
	to << M->rows << "\n";
	int i, j;
	for(i = 0; i < M->rows; i++) {
		for(j = 0; j < M->cols; j++)
			to << M->a[i][j] << "\n";
	};
};

void putsep(int v) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	scheme *X = (scheme *)var[v].var_ptr;

	if(X->sep == 0)
		sep(*X);
	else
		cout << Line;

	to << 1 << "\n";
	to << X->s << "\n";
	for(int i = 0; i < X->s; i++) {
		X->separator->gen[i].put(to);
		to << "\n";
	};
};

void putGroebner(int v) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	scheme *X = (scheme *)var[v].var_ptr;

	if(X->delta == 0)
		bu_mo(*X);
	else
		cout << Line;

	to << 1 << "\n";
	to << X->Ix->no_of_gens << "\n";
	for(int i = 0; i < X->Ix->no_of_gens; i++) {
		X->Ix->gen[i].put(to);
		to << "\n";
	};
};

void putmin(int v) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	scheme *X = (scheme *)var[v].var_ptr;

	if(X->min_Ix == 0)
		min_Ix(*X);
	else
		cout << Line;

	to << 1 << "\n";
	to << X->min_Ix->no_of_gens << "\n";
	for(int i = 0; i < X->min_Ix->no_of_gens; i++) {
		X->min_Ix->gen[i].put(to);
		to << "\n";
	};
};

void putA_d(int v) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	scheme *X = (scheme *)var[v].var_ptr;

	if(X->no_of_pr == 0)
		omega(*X);
	else
		cout << Line;

	to << X->no_of_pr << "\n";
	to << X->A_d_cols << "\n";
	int i, n;
	for(i = 0; i < X->A_d_cols; i++)
		for(n = 0; n < X->no_of_pr; n++) {
			X->A_d[i + n * X->A_d_cols].put(to);
			to << "\n";
		};
};

/*******************************************************************
cyclotom
*******************************************************************/

void show_settings(void);

void cyclotom(void) {
#ifdef field_QN
	if(dim != 2) {
		cerr << "dimension = 2 required !!! \n";
		return;
	};
	int a, i, j, v, s;
	char cc[13];

	cin.width(13);
	cin >> cc;

	for (i = 0; i < strlen(cc); i++)
		if (! isalnum((int) cc[i])) {
			error(105);
			return;
		};

	read_var(v, cc);
	if(v != -2) {
		error(111);  // par name already exists
		return;
	};

	s = 0;
	cin >> s;
	if(s <= 0) {
		cout << "first parameter has to be greater than zero !\n";
		return;
	};

	a = 0;
	cin >> a;
	if(a <= 0 || a >= s) {
		cout << "second parameter has to be greater than zero and less than the first parameter!\n";
		return;
	};

	for (i = 0; no_of_vars; i++) {
		if (var[i].name[0] == '\0') {
			strcpy(var[i].name, cc);
			var[i].typ  = 3;
			v = i;
			break;
		};
	};

	if (i == no_of_vars) {
		error(110);
		return;
	};

	set(s);

	var[v].var_ptr = new scheme(s);
	for(i = 0; i < s; i++) {
		((scheme *)var[v].var_ptr)->P->a[i][0] = 0;
		if(i < QN::n)
			((scheme *)var[v].var_ptr)->P->a[i][0].a[i] = 1;
		else
			((scheme *)var[v].var_ptr)->P->a[i][0] = QN_phi[i-QN::n];
	};
	for(i = 0; i < s; i++) {
		((scheme *)var[v].var_ptr)->P->a[i][1] =
		((scheme *)var[v].var_ptr)->P->a[i][0];
		for(j = 0; j < a - 1; j++)
			((scheme *)var[v].var_ptr)->P->a[i][1] *=
			((scheme *)var[v].var_ptr)->P->a[i][0];
	};
	show_settings();
#else
	cout << "please use 'COP_QN' for computing cyclotomic schemes\n" << Line;
#endif
};

void cyclotom2(void) {
#ifdef field_QN
	if(dim != 2) {
		cerr << "dimension = 2 required !!! \n";
		return;
	};
	int a, b, i, j, v, s;
	char cc[13];

	cin.width(13);
	cin >> cc;

	for (i = 0; i < strlen(cc); i++)
		if (! isalnum((int) cc[i])) {
			error(105);
			return;
		};

	read_var(v, cc);
	if(v != -2) {
		error(111);  // par name already exists
		return;
	};

	s = 0;
	cin >> s;
	if(s <= 0) {
		cout << "first parameter has to be greater than zero !\n";
		return;
	};

	a = 0;
	cin >> a;
	if(a <= 0 || a >= s) {
		cout << "second parameter has to be greater than zero and less than the first parameter!\n";
		return;
	};

	b = 0;
	cin >> b;
	if(b <= 0 || b >= a) {
		cout << "third parameter has to be greater than zero and less than the second parameter!\n";
		return;
	};

	for (i = 0; no_of_vars; i++) {
		if (var[i].name[0] == '\0') {
			strcpy(var[i].name, cc);
			var[i].typ  = 3;
			v = i;
			break;
		};
	};

	if (i == no_of_vars) {
		error(110);
		return;
	};

	set(s);

	var[v].var_ptr = new scheme(s);
	for(i = 0; i < s; i++) {
		((scheme *)var[v].var_ptr)->P->a[i][0] = 0;
		if(i < QN::n)
			((scheme *)var[v].var_ptr)->P->a[i][0].a[i] = 1;
		else
			((scheme *)var[v].var_ptr)->P->a[i][0] = QN_phi[i-QN::n];
	};
	for(i = 0; i < s; i++) {
		((scheme *)var[v].var_ptr)->P->a[i][1] =
		((scheme *)var[v].var_ptr)->P->a[i][0];
		for(j = 0; j < a - 1; j++)
			((scheme *)var[v].var_ptr)->P->a[i][1] *=
			((scheme *)var[v].var_ptr)->P->a[i][0];
	};
	QN tmp(0);
	for(i = 0; i < s; i++) {
		tmp = ((scheme *)var[v].var_ptr)->P->a[i][0];
		for(j = 0; j < b - 1; j++)
			tmp *= ((scheme *)var[v].var_ptr)->P->a[i][0];
		((scheme *)var[v].var_ptr)->P->a[i][1] += tmp;
	};
	show_settings();
#else
	cout << "please use 'COP_QN' for computing cyclotomic schemes\n" << Line;
#endif
};

void gencycl(void) {
#ifdef field_QN
	if(dim != 2) {
		cerr << "dimension = 2 required !!! \n";
		return;
	};
	int i, j, k, v, s;
	char cc[13];

	cin.width(13);
	cin >> cc;

	for (i = 0; i < strlen(cc); i++)
		if (! isalnum((int) cc[i])) {
			error(105);
			return;
		};

	read_var(v, cc);
	if(v != -2) {
		error(111);  // par name already exists
		return;
	};

	s = 0;
	cin >> s;
	if(s <= 0) {
		cout << "first parameter has to be greater than zero !\n";
		return;
	};

	for (i = 0; no_of_vars; i++) {
		if (var[i].name[0] == '\0') {
			strcpy(var[i].name, cc);
			var[i].typ  = 3;
			v = i;
			break;
		};
	};

	if (i == no_of_vars) {
		error(110);
		return;
	};

	set(s);

	var[v].var_ptr = new scheme(s);
	for(i = 0; i < s; i++) {
		((scheme *)var[v].var_ptr)->P->a[i][0] = 0;
		if(i < QN::n)
			((scheme *)var[v].var_ptr)->P->a[i][0].a[i] = 1;
		else
			((scheme *)var[v].var_ptr)->P->a[i][0] = QN_phi[i-QN::n];
	};

	QN tmp(0);
	Rational tmp1(0);
	cout << "enter coefficients :\n";
	for(i = s-1; i >= 0; i--) {
		cout << "c_" << i << " : ";
		cin >> tmp1;
		for(k = 0; k < s; k++) {
			tmp = ((scheme *)var[v].var_ptr)->P->a[k][0];
			for(j = 0; j < i - 1; j++)
				tmp *= ((scheme *)var[v].var_ptr)->P->a[k][0];
			tmp *= tmp1;
			((scheme *)var[v].var_ptr)->P->a[k][1] += tmp;
		};
	};
	show_settings();
#else
	cout << "please use 'COP_QN' for computing cyclotomic schemes\n" << Line;
#endif
};

/*******************************************************************
random
*******************************************************************/

void random1(void) {
	int i, j, k, v, s;
	char cc[13];

	cin.width(13);
	cin >> cc;

	for (i = 0; i < strlen(cc); i++)
		if (! isalnum((int) cc[i])) {
			error(105);
			return;
		};

	read_var(v, cc);
	if(v != -2) {
		error(111);  // par name already exists
		return;
	};

	for (i = 0; no_of_vars; i++) {
		if (var[i].name[0] == '\0') {
			strcpy(var[i].name, cc);
			var[i].typ  = 3;
			v = i;
			break;
		};
	};

	if (i == no_of_vars) {
		error(110);
		return;
	};

	s = 0;
	cin >> dec >> s;
	if(s <= 0) {
		cout << "number of points has to be greater than zero !\n";
		var[v].name[0] = '\0';
		return;
	};

	RandomInteger rnd(range1, range2, &gen);

	var[v].var_ptr = new scheme(s);
	for(i = 0; i < s; i++)
		for(j = 0; j < dim; j++)
#ifdef field_QN
			for(k = 0; k < QN::n; k++)
				((scheme *)var[v].var_ptr)->P->a[i][j].a[k] = rnd();
#else
			((scheme *)var[v].var_ptr)->P->a[i][j] = rnd();
#endif
	// is reduced ?
	for(i = 0; i < s; i++) {
		for(j = i + 1; j < s; j++) {
			for(k = 0; k < dim; k++)
				if(((scheme *)var[v].var_ptr)->P->a[i][k] !=
					((scheme *)var[v].var_ptr)->P->a[j][k])
					break;
			if(k == dim) {
				for(k = 0; k < dim; k++)
					((scheme *)var[v].var_ptr)->P->a[j][k] = rnd();
				i = 0;
				j = 0;
			};
		};
	};
	cout << Line;
};

void random2(void) {
	int i, j, k, v, s;
	char cc[13];

	cin.width(13);
	cin >> cc;

	for (i = 0; i < strlen(cc); i++)
		if (! isalnum((int) cc[i])) {
			error(105);
			return;
		};

	read_var(v, cc);
	if(v != -2) {
		error(111);  // par name already exists
		return;
	};

	for (i = 0; no_of_vars; i++) {
		if (var[i].name[0] == '\0') {
			strcpy(var[i].name, cc);
			var[i].typ  = 3;
			v = i;
			break;
		};
	};

	if (i == no_of_vars) {
		error(110);
		return;
	};

	s = 0;
	cin >> s;
	if(s <= 0) {
		cout << "number of points has to be greater than zero !\n";
		var[v].name[0] = '\0';
		return;
	};

	RandomInteger rnd(range1, range2, &gen);

	var[v].var_ptr = new scheme(s);
	for(j = 0; j < dim; j++)
		((scheme *)var[v].var_ptr)->P->a[0][j] = 0;
	for(i = 0; i < min(dim,s-1); i++)
		for(j = 0; j < dim; j++)
			if(i == j) ((scheme *)var[v].var_ptr)->P->a[i+1][j] = 1;
			else ((scheme *)var[v].var_ptr)->P->a[i+1][j] = 0;

	for(i = dim +1; i < s; i++)
		for(j = 0; j < dim; j++)
#ifdef field_QN
			for(k = 0; k < QN::n; k++)
				((scheme *)var[v].var_ptr)->P->a[i][j].a[k] = rnd();
#else
			((scheme *)var[v].var_ptr)->P->a[i][j] = rnd();
#endif
	// is reduced ?
	for(i = 0; i < s; i++) {
		for(j = i + 1; j < s; j++) {
			for(k = 0; k < dim; k++)
				if(((scheme *)var[v].var_ptr)->P->a[i][k] !=
					((scheme *)var[v].var_ptr)->P->a[j][k])
					break;
			if(k == dim) {
				for(k = 0; k < dim; k++)
					((scheme *)var[v].var_ptr)->P->a[j][k] = rnd();
				i = 0;
				j = 0;
			};
		};
	};
	cout << Line;
};

/*******************************************************************
lingenpos
*******************************************************************/

void lingenpos(scheme &X) {
	int i;
	int *c, *r;
	int d = min(X.s, dim+1);
	c = new int[d + 1];
	r = new int[d + 1];
	for(i = 0; i < d + 1; i++)
		c[i] = r[i] = i;

	r[d] = max(X.s, dim+1);
	int quit = 0;
	K m;

	// calc minors
	while (!quit) {
		if(d == X.s)
			m = sign(X.P->minor(d, r, c));
		else
			m = sign(X.P->minor(d, c, r));
		if(sign(m) != 0) break;
		for(i = 0; i < d; i++) {
			r[i]++;
			if (r[i] == r[i + 1]) {
				r[i] = i;
				if(i == d - 1) {
					quit = 1;
					break;
				};
			}
			else break;
		};
	};
	if(quit == 1) cout << "scheme is not in linearly general position\n";
	else cout << "scheme is in linearly general position\n";
	cout << Line;
};

/*******************************************************************
del
*******************************************************************/

void del(int v, int mode = 0) {

	if (var[v].typ == 3)
		delete (scheme *) var[v].var_ptr;
	if (!mode) // quiet mode
		cout << var[v].name << " deleted\n";

	var[v].name[0] = '\0';
	var[v].typ = 0;
};

void del_all(int mode = 0) {
	int flag = 0;
	for(int i = 0; i < no_of_vars; i++) {
		if (var[i].name[0] != '\0') {
			del(i, mode);
			flag = 1;
		};
	};
	if(!mode & flag) cout << Line;
};

/*******************************************************************
display
*******************************************************************/

void display(int v, int mode = 0) {
	switch (var[v].typ) {
		case 3: // scheme
			cout << "scheme " << var[v].name;
			break;
	};

	if (mode) { // full mode
		cout << ":\n";
		switch (var[v].typ){
			case 3: // scheme
				cout << *(scheme*) var[v].var_ptr;
				break;
		};
	}
	else cout << "\n";
};

void list_all(int mode = 0) {
	for(int i = 0; i < no_of_vars; i++)
		if (var[i].name[0] != '\0')
			display(i, mode);
};

void write(int v) {
	cout << "enter filename : ";
	char file_name[100];
	cin >> file_name;
	ofstream to(file_name);
	if(!to)  {
		error(30);
		return;
	};
	to << *((scheme *)var[v].var_ptr);
	cout << Line;
};

/*******************************************************************
miscellaneous
*******************************************************************/

void show_settings(void) {
	cout << "dimension         : " << dim << "\n";
	cout << "ring              : ";
#ifdef field_QN
	cout << "Q()";
#endif
#ifdef field_Q
	cout << "Q";
#endif
#ifdef field_Fp
	cout << "F" << Fp::prime;
#endif
#ifdef field_Fp8
	cout << "F" << Fp8::prime;
#endif
	cout << "[";
	for(int i = 0; i < proj; i++)
		cout << varstring[i];
	cout << "]\n";
#ifdef field_QN
	cout << "                     = " << QN::N << "-th root of unity\n";
	cout << "                    a Q-basis for Q() is ^0,...,^" << QN::n-1 << "\n";
#endif
	cout << "monomial ordering : ";
	switch (monom_ordering) {
		case 0 :
			cout << "deg-lex\n";
			break;
		case 1 :
			cout << "deg-rev-lex\n";
			break;
	};
	cout << Line;
};

void identifiers(void) {
	cout << "enter " << proj << " identifiers for current ring : ";
	cin.width(27);
	cin >> varstring;
	if(strlen(varstring) < proj) {
		cerr << "not enough indentifiers\n";
		identifiers();
	};
};

void set_dim() {
	int d;
	cin >> d;
	if(d <= 1) {
		cout << "\ndimension has to be greater than one !!!\n";
		return;
	};
	dim = d;
	del_all();
	proj = dim + 1;
	if(strlen(varstring) < proj) {
		cerr << "not enough indentifiers\n";
		identifiers();
	};
	show_settings();
};

void root() {
#ifdef field_QN
	int d;
	cin >> d;
	if(d <= 1) {
		cout << "\nroots of N with N > 1 !!!\n";
		return;
	};
	set(d);
	del_all();
	show_settings();
#else
	cout << "please use 'COP_QN' for computations over the field with n-th roots of unity\n" << Line;
#endif
};

void cyclpol() {
#ifdef field_QN
	int d = dim;
	dim = 1;
	polynomial f,g;
	monom m;
	f = 0;
	m = 0;
	for(int i = 0; i < QN::n; i++) {
		if(sign(QN_phi[0].a[i]) == 0) continue;
		m.exp[0] = i;
		g = m;
		g.coef[0] = QN_phi[0].a[i];
		f -= g;
	};
	m.exp[0] = QN::n;
	g = m;
	g.coef[0] = 1;
	f += g;
	counter = 0;
	cout << "the " << QN::N << "-th cyclotomic polynomial over Q is :\n";
	cout << f << "\n" << Line;
	dim = d;
#else
	cout << "please use 'COP_QN' for computations over the field with n-th roots of unity\n" << Line;
#endif
};

void set_defaults() {
	dim  = 3;
	proj = 4;
	verbose = 1;
	LL = 80;
	Line[LL] = '\0';
	Line[LL - 1] = '\n';
#ifdef field_Fp
	range1 = 0;
	range2 = min(50, Fp::prime - 1);
#else
#ifdef field_Fp8
	range1 = 0;
	range2 = min(50, Fp8::prime - 1);
	Fp8 x;
	x.init();
#else
	range1 = -50;
	range2 = 50;
#endif
#endif
// set default ordering -> deg rev lex
	monom_ordering_lt = degrevlex_lt;
	monom_ordering_le = degrevlex_le;
	proj_monom_ordering_lt = proj_degrevlex_lt;
	proj_monom_ordering_le = proj_degrevlex_le;
	monom_ordering = 1; // deg rev lex
#ifdef field_QN
	QN::mat[0] = new Rational[1];
	set(5);
#endif
};

void hello() {
	cout << Line;
	cout << "Welcome to C O P  ---  C omputation   O f   P oints  (C) 1992-1994\n";
	cout << Line;
	cout << "developed by Stefan Beck & Martin Kreuzer at the University of Regensburg\n";
	cout << Line;
};

void deglex(void) {
	monom_ordering = 0;
	monom_ordering_lt = deglex_lt;
	monom_ordering_le = deglex_le;
	proj_monom_ordering_lt = proj_deglex_lt;
	proj_monom_ordering_le = proj_deglex_le;
	show_settings();
	del_all();
};

void degrevlex(void) {
	monom_ordering = 1;
	monom_ordering_lt = degrevlex_lt;
	monom_ordering_le = degrevlex_le;
	proj_monom_ordering_lt = proj_degrevlex_lt;
	proj_monom_ordering_le = proj_degrevlex_le;
	show_settings();
	del_all();
};

void characteristic(void) {
#ifdef field_Fp
	int p;
	cin >> p;
	if(p <= 1 || p > 32003) {
		cout << "characteristic has to be a prime (2 - 32003) !\n";
		return;
	};
	Fp::prime = p;
	range2 = min(range2, Fp::prime - 1);
	cout << Line;
	show_settings();
	del_all();
#else
#ifdef field_Fp8
	int p;
	cin >> p;
	if(p <= 1 || p > 251) {
		cout << "characteristic has to be a prime (2 - 251) !\n";
		return;
	};
	Fp8::prime = p;
	Fp8 x;
	x.init();
	range2 = min(range2, Fp8::prime - 1);
	cout << Line;
	show_settings();
	del_all();
#else
	cout << "please use 'COP_Fp' for computations over finite fields \n" << Line;
#endif
#endif
};

void help(void) {
	cout << "list of commands in alphabetical order :";
	int j, n = LL / 13;
	for(int i = 0; i < no_of_cmds; i++) {
		if(i % n == 0) cout << "\n";
		cout << command[i].name;
		for(j = 0; j < 13 - strlen(command[i].name); j++) cout << " ";
	};
	cout << "\n" << Line;
};

void range() {
	cout << "current range : [" << range1 << "," << range2 << "]\n";
#ifdef field_Q
	cout << "enter lower border : ";
	cin >> range1;
#endif
#ifdef field_QN
	cout << "enter lower border : ";
	cin >> range1;
#endif
	range2 = range1 - 1;
	while(range1 >= range2) {
		cout << "enter upper border : ";
		cin >> range2;
#ifdef field_Fp
		range2 = min(range2, Fp::prime - 1);
#endif
#ifdef field_Fp8
		range2 = min(range2, Fp8::prime - 1);
#endif
		if(range1 >= range2)
			cout << "upper border must be greater than lower border !!!\n";
	};
	cout << "current range : [" << range1 << "," << range2 << "]\n";
	cout << Line;
};

/*******************************************************************
parser
*******************************************************************/

int quit;

void read_var(int &v, char *cc) {
	for (int i = 0; i < no_of_vars; i++)
		if (! strncmp(cc, var[i].name, 12)) {  // var found
			v = i;
			return;
		};
	v = -2;
};

void execute(int cmd, int var1);

void parser() {
	char c;
	char cc[13];
	int cmd, var1;
	int i;

	while (quit == 0) {
		// reset
		cmd = -1;
		var1 = -1;

		// prompt
		cout << "% ";

		cin.width(13);
		cin >> cc;

		for (i = 0; i < no_of_cmds; i++)
			if (! strncmp(cc, command[i].name, 12)) { // cmd found
				cmd = i;
				break;
			};

		if (cmd >= 0) {
			if (command[cmd].par1_typ >= 0) {
				if(cin.peek() != '\n') {
					cin.width(13);
					cin >> cc;
					read_var(var1, cc);
				};
			};
			execute(cmd, var1);
		}
		else {
			read_var(var1, cc);
			if (var1 >= 0)
				execute(14, var1); // display var1
			else
				error(102);          // cmd or var unknown
		};
		// what now is in the istream has to be ignored
		while(cin.peek() != '\n') {
			cin.get(c);
			if(!isspace(c)) {
				cout << "unexpected '" << c;
				while(cin.peek() != '\n') {
					cin.get(c);
					cout << c;
				};
				cout << "'\n";
			};
		};
	};
};

/*******************************************************************
execute
*******************************************************************/

void execute(int cmd, int var1) {
	// check parameters
	if (command[cmd].par1_typ > 0) {
		if (var1 == -2) {
			error(107); // first par name unknown
			return;
		}
		else if (var1 == -1) {
			error(104); // parameter expected after cmd
			return;
		}
		else if (var[var1].typ != command[cmd].par1_typ) {
			error(108); // wrong type
			return;
		};
	};

	if (command[cmd].par1_typ == 0) {
		if (var1 == -2) {
			error(107); // first par name unknown
			return;
		}
		else if (var1 == -1) {
			error(104); // parameter expected after cmd
			return;
		};
	};

	char file_name[100];

	switch (cmd) {
		case 0 : // alpha
			alpha_X(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 1 : // betamat
			disp_beta_matrix(*(scheme*)var[var1].var_ptr);
			break;

		case 2 : // bumo
			bu_mo(*(scheme*)var[var1].var_ptr);
			break;

		case 3 : // char
			characteristic();
			break;

		case 4 : // cyclotom
			cyclotom();
			break;

		case 5 : // cyclotom2
			cyclotom2();
			break;

		case 6 : // cyclpol
			cyclpol();
			break;

		case 7 : // deglex
			deglex();
			break;

		case 8 : // degrevlex
			degrevlex();
			break;

		case 9 : // del
			del(var1);
			break;

		case 10 : // del_all
			del_all();
			break;

		case 11 : // delta
			delta_X(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 12 : // deltafn
			delta_fct(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 13 : // dim
			set_dim();
			break;

		case 14 : // display
			display(var1, 1);
			break;

		case 15 : // display_all
			list_all(1);
			break;

		case 16 : // exit
			quit = 1;
			break;

		case 17 : // gBasis
			groebner(*(scheme*)var[var1].var_ptr);
			break;

		case 18 : // gencycl
			gencycl();
			break;

		case 19 : // get
			get_scheme();
			break;

		case 20 : // help
			help();
			break;

		case 21 : // hilbertfn
			hilb_fct(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 22 : // identifiers
			identifiers();
			show_settings();
			break;

		case 23 : // is_CB
			is_CB(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 24 : // is_duf
			is_delta_uf(*(scheme*)var[var1].var_ptr);
			break;

		case 25 : // is_gor
			is_ar_Go(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 26 : // is_uf
			is_n_uf(*(scheme*)var[var1].var_ptr);
			break;

		case 27 : // lingenpos
			lingenpos(*(scheme*)var[var1].var_ptr);
			break;

		case 28 : // linid
			linid(*(scheme*)var[var1].var_ptr);
			disp_linid(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 29 : // linom
			linom(*(scheme*)var[var1].var_ptr);
			disp_linom(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 30 : // linomsyz
			linomsyz(*(scheme*)var[var1].var_ptr);
			break;

		case 31 : // linres
			linid(*(scheme*)var[var1].var_ptr);
			linom(*(scheme*)var[var1].var_ptr);
			disp_linres(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 32 : // linsyz
			linsyz(*(scheme*)var[var1].var_ptr);
			break;

		case 33 : // list
			list_all();
			break;

		case 34 : // min
			disp_min(*(scheme*)var[var1].var_ptr);
			break;

		case 35 : // mindegs
			deg_min(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 36 : // omega
			disp_omega(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 37 : // omegamat
			disp_omegamat(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 38 : // orderideal
			order_ideal(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 39 : // putbeta
			putbeta(var1);
			break;

		case 40 : // putgbasis
			putGroebner(var1);
			break;

		case 41 : // putlinid
			cout << "enter filename : ";
			cin >> file_name;
			putlinid(*(scheme*)var[var1].var_ptr, file_name, ios::out);
			break;

		case 42 : // putlinom
			cout << "enter filename : ";
			cin >> file_name;
			putlinom(*(scheme*)var[var1].var_ptr, file_name, ios::out);
			break;

		case 43 : // putlinomsyz
			putlinomsyz(*(scheme*)var[var1].var_ptr);
			break;

		case 44 : // putlinres
			cout << "enter filename : ";
			cin >> file_name;
			putlinid(*(scheme*)var[var1].var_ptr, file_name, ios::out);
			putlinom(*(scheme*)var[var1].var_ptr, file_name, ios::app);
			break;

		case 45 : // putlinsyz
			putlinsyz(*(scheme*)var[var1].var_ptr);
			break;

		case 46 : // putmin
			putmin(var1);
			break;

		case 47 : // putomega
			putA_d(var1);
			break;

		case 48 : // putscheme
			putscheme(var1);
			break;

		case 49 : // putsep
			putsep(var1);
			break;

		case 50 : // quiet
			verbose = 0;
			break;

		case 51 : // quit
			quit = 1;
			break;

		case 52 : // random
			random1();
			break;

		case 53 : // random2
			random2();
			break;

		case 54 : // range
			range();
			break;

		case 55 : // ring
			show_settings();
			break;

		case 56 : // root
			root();
			break;

		case 57 : // scheme
			create(3);
			break;

		case 58 : // sep
			separators(*(scheme*)var[var1].var_ptr);
			break;

		case 59 : // sigma
			sigma_X(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 60 : // socle
			disp_socle(cout, *(scheme*)var[var1].var_ptr);
			break;

		case 61 : // typ
			typ(*(scheme*)var[var1].var_ptr);
			break;

		case 62 : // verbose
			verbose = 1;
			break;

		case 63 : // write
			write(var1);
			break;

		case 64 : // writelinid
			writelinid(*(scheme*)var[var1].var_ptr);
			break;

		case 65 : // writelinom
			writelinom(*(scheme*)var[var1].var_ptr);
			break;

		case 66 : // writelinres
			writelinres(*(scheme*)var[var1].var_ptr);
			break;

	};
};

//              S T A R T
int main() {
	set_new_handler(mem_warn);

	set_defaults();
	hello();
	show_settings();
	parser();

	del_all(1);

	cout << Line << "thanks for using COP\n" << Line;

	return 0;
};
